"""
Module for generating samples from :abbr:`SVVJ(Stochastic Volatility with
jumps in the Variance)`

Note that we are generating i.i.d. samples with fixed conditions instead of
a trajectory of samples!
"""
import math
import numpy as np

rng = np.random.default_rng()


def IZ(t, jumpsize, jumptime):
    f = 0
    if len(jumpsize) == 0: return f
    i = 0
    while i < len(jumpsize) and jumptime[i] <= t:
        f += jumpsize[i]
        i += 1
    return f


def IEZ(t, jumpsize, jumptime, k):
    f = 0
    if len(jumpsize) == 0: return f
    i = 0
    while i < len(jumpsize) and jumptime[i] <= t:
        f += math.exp(k * jumptime[i]) * jumpsize[i]
        i += 1
    return f


def incIEII(v0, t0, dt, k, theta, sigma):
    # IE = \int_0^t e^{ks} \sqrt{v_s} dw_s^v
    # I =  \int_0^t        \sqrt{v_s} dw_s^v
    # I2 = \int_0^t        \sqrt{v_s} dw_s
    # dv_t = k(theta - v_t)dt + sigma * sqrt(v_t)dw_t^v + dz_t
    eps_v = math.sqrt(dt) * rng.normal(0, 1)
    eps_y = math.sqrt(dt) * rng.normal(0, 1)
    #
    vt = v0 + k * (theta - v0) * dt + sigma * math.sqrt(v0) * eps_v
    #
    if vt < 0: vt = 0
    #
    I = math.sqrt(v0) * eps_v
    IE = math.exp(k * t0) * math.sqrt(v0) * eps_v
    I2 = math.sqrt(v0) * eps_y
    return IE, I, I2, vt


def rIEII(v0, k, theta, sigma, h, jumpsize, jumptime, n_segment):
    IE, I, I2 = 0, 0, 0
    dt = h / n_segment
    #
    if len(jumptime) == 0:
        for i in range(n_segment):
            t0 = i * dt
            incIE, incI, incI2, vt = incIEII(v0, t0, dt, k, theta, sigma)
            IE += incIE
            I += incI
            I2 += incI2
            v0 = vt
        return IE, I, I2
    # at least one jump
    j = 0
    for i in range(n_segment):
        t0 = i * dt
        t = (i + 1) * dt
        while j < len(jumptime) and jumptime[j] <= t:
            # print(f"within {i+1}-th segment: [{i*dt}, {(i+1)*dt})")
            # print(f"the {j+1}-th jump, jump size = {jumpsize[j]}, jump time = {jumptime[j]}")
            # pre jump
            dt1 = jumptime[j] - t0
            incIE, incI, incI2, vt = incIEII(v0, t0, dt1, k, theta, sigma)
            IE += incIE
            I += incI
            I2 += incI2
            # jump
            vt += jumpsize[j]
            #
            t0 = jumptime[j]
            v0 = vt
            j += 1
        if t0 < t:
            dt1 = t - t0
            incIE, incI, incI2, vt = incIEII(v0, t0, dt1, k, theta, sigma)
            IE += incIE
            I += incI
            I2 += incI2
            #
            v0 = vt
    return IE, I, I2


def rSVVJ(v0, mu, k, theta, sigma, rho, T, jumpsize, jumptime, n, n_segment):
    r"""Generate i.i.d. Random samples from mdl_svvj by Euler approximation.

    :param float v0: value of the initial variance :math:`v(0)`
    :param float mu: parameter :math:`\mu`
    :param float k: parameter :math:`k`
    :param float theta: parameter :math:`\theta`
    :param float sigma: parameter :math:`\sigma_v`
    :param float rho: parameter :math:`\rho`
    :param float T: end time
    :param tuple jumpsize: jump sizes over [0, T]
    :param tuple jumptime: jump time points over [0, T]
    :param int n: number of samples to generate
    :param int n_segment: number of segments to split [0, T]
    :return: i.i.d. samples
    :rtype: numpy.ndarray
    """
    y = np.empty(n)
    beta = (1 - math.exp(-k * T)) / (2 * k)
    c = (mu - theta / 2) * T - (v0 - theta) * beta
    c += math.exp(-k * T) * IEZ(T, jumpsize, jumptime, k) / (2 * k)
    c -= IZ(T, jumpsize, jumptime) / (2 * k)
    #
    for i in range(n):
        IE, I, I2 = rIEII(v0, k, theta, sigma, T, jumpsize, jumptime, n_segment)
        y[i] = c + math.exp(-k * T) * IE * sigma / (2 * k)
        y[i] += (rho - sigma / (2 * k)) * I
        y[i] += math.sqrt(1 - rho ** 2) * I2
    return y


if __name__ == "__main__":
    import sys

    v0, h = 0.007569, 1
    mu, k, theta, sigma, rho = 0.0319, 3.46, 0.008, 0.14, -0.82
    lmbd, mu_v = 0.47, 0.05
    #
    args = sys.argv[1:]
    numJ = rng.poisson(lmbd * h) if len(args) == 0 else int(args[0])
    jumpsize = rng.exponential(mu_v, numJ)
    jumptime = rng.uniform(0, h, numJ)
    jumptime = sorted(jumptime)
    # change to tuples
    jumpsize, jumptime = tuple(jumpsize), tuple(jumptime)
    #
    n, n_segment = 20, 10
    y = rSVVJ(v0, mu, k, theta, sigma, rho, h, jumpsize, jumptime, n, n_segment)
    print(f"\nGiven v0 = {v0} and {numJ} jumps, ")
    print(f"jumpsize = {jumpsize}, \njumptime = {jumptime}")
    print(f"\n{n} i.i.d. samples from SVVJ: ")
    print(y)



"""
Central Moments for Two-Factor SV model
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.ito_mom import moment_v
from ajdmom.itos_mom import moment_IEI_IEII, t_mul_t0
from ajdmom.utils import comb


def b_n(n1, n2, n3, n4, n5, n6, n7):
    """Coefficient corresponding to the combination

    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{1,n-1}` being selected.
    :param int n3: times of :math:`v_{2,n-1}` being selected.
    :param int n4: times of :math:`eI_{1,n}` being selected.
    :param int n5: times of :math:`I_{1,n}` being selected.
    :param int n6: times of :math:`eI_{2,n}` being selected.
    :param int n7: times of :math:`I_{2,n}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('e^{-k1h}','k1^{-}','theta1','sigma_v1',
       'e^{-k2h}','k2^{-}','theta2','sigma_v2').
    :rtype: Poly
    """
    b = Poly()
    kf = ['e^{-k1h}', 'k1^{-}', 'theta1', 'sigma_v1',
          'e^{-k2h}', 'k2^{-}', 'theta2', 'sigma_v2']
    b.set_keyfor(kf)
    #
    for i1 in range(n1, -1, -1):  # i1: 1/k1 * theta1
        for i2 in range(n1 - i1, -1, -1):  # i2: -e^{-k1h}/k1 * theta1
            for i3 in range(n1 - i1 - i2, -1, -1):  # i3: 1/k2 * theta2
                i4 = n1 - i1 - i2 - i3  # i4: -e^{-k2h}/k2 * theta2
                for j in range(n2 + 1):  # j: -e^{-k1h}/k1
                    for q in range(n3 + 1):  # q: -e^{-k2h}/k2
                        num = math.comb(n1, i1)
                        num *= math.comb(n1 - i1, i2)
                        num *= math.comb(n1 - i1 - i2, i3)
                        num *= math.comb(n2, j)
                        num *= math.comb(n3, q)
                        key = (i2 + j, i1 + i2 + n2 + n4 + n5, i1 + i2, n4 + n5,
                               i4 + q, i3 + i4 + n3 + n6 + n7, i3 + i4, n6 + n7)
                        num *= (-1) ** (i2 + i4 + (n2 + j) + (n3 + q) + n5 + n7)
                        den = 2 ** (n1 + n2 + n3 + n4 + n5 + n6 + n7)  # l - n8
                        b.add_keyval(key, Frac(num, den))
    return b


def vvee_IEI_IEII(n2, n3, n4, n5, n6, n7, n8):
    r"""Multiply :math:`IEI\_IEII` by :math:`v_{1,n-1}v_{2,n-1}e^{-k_1nh}e^{-k_2nh}`

    :math:`E[v_{1,n-1}^{n_2}v_{2,n-1}^{n_3}e^{-n_4k_1 nh}e^{-n_6k_2 nh} IEI\_IEII]`
       where

       .. math::

          IEI\_IEII = E[eI_{1,n}^{n_4} I_{1,n}^{n_5} eI_{2,n}^{n_6} I_{2,n}^{n_7}
          I_{n}^{*n_8}|v_{1,n-1}v_{2,n-1}]

    :param int n2: times of :math:`v_{1,n-1}` being selected.
    :param int n3: times of :math:`v_{2,n-1}` being selected.
    :param int n4: times of :math:`eI_{1,n}` being selected.
    :param int n5: times of :math:`I_{1,n}` being selected.
    :param int n6: times of :math:`eI_{2,n}` being selected.
    :param int n7: times of :math:`I_{2,n}` being selected.
    :param int n8: times of :math:`I_{n}^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poln = Poly()
    kf = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2')
    poln.set_keyfor(kf)
    #
    poly = moment_IEI_IEII(n4, n5, n6, n7, n8)
    # keyfor =
    # ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    # 'e^{(m_4*k1+m_6*k2)(n-1)h}','e^{(j_1*k1+j_2*k2)[t-(n-1)h]}','[t-(n-1)h]',
    # 'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2')
    for k in poly:
        # * e^{-n4*k1*nh} e^{-n6*k2*nh}  * v_{1,n-1}^{n2} v_{2,n-1}^{n3}
        t1 = (n4 - k[2][0], n6 - k[2][1])
        key = (k[0], t1, k[3], k[4] + n2, k[5], k[6], k[7] + n3, k[8], k[9])
        poln.add_keyval(key, poly[k])
    return poln


def moment_comb(n, n1, n2, n3, n4, n5, n6, n7, n8):
    r"""Moment for this combination of given n1 to n8 with n.

    :param int n: l in :math:`E[\overline{y}_{n}^l]`.
    :param int n1: times of constant being selected.
    :param int n2: times of :math:`v_{1,n-1}` being selected.
    :param int n3: times of :math:`v_{2,n-1}` being selected.
    :param int n4: times of :math:`eI_{1,n}` being selected.
    :param int n5: times of :math:`I_{1,n}` being selected.
    :param int n6: times of :math:`eI_{2,n}` being selected.
    :param int n7: times of :math:`I_{2,n}` being selected.
    :param int n8: times of :math:`I_{n}^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :rtype: Poly
    """
    poly = vvee_IEI_IEII(n2, n3, n4, n5, n6, n7, n8)
    # ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
    # 'e^{-(n1*k1+n2*k2)h}','h', 'v_{1,n-1}','theta1','sigma_v1',
    # 'v_{2,n-1}','theta2','sigma_v2')
    b = b_n(n1, n2, n3, n4, n5, n6, n7)
    # ['e^{-k1h}','k1^{-}','theta1','sigma_v1',
    #  'e^{-k2h}','k2^{-}','theta2','sigma_v2']
    kf = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2')
    #
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k1 in poly:
        for k2 in b:
            # t0 = list(k1[0]); t0.insert(0,(1,0,k2[1])); t0.insert(0,(0,1,k2[5]))
            # t0 = tuple(t0)
            t0 = t_mul_t0((1, 0, k2[1]), k1[0])
            t0 = t_mul_t0((0, 1, k2[5]), t0)
            #
            t1 = (k1[1][0] + k2[0], k1[1][1] + k2[4])
            #
            key = (t0, t1, k1[2], k1[3], k1[4] + k2[2], k1[5] + k2[3],
                   k1[6], k1[7] + k2[6], k1[8] + k2[7])
            val = poly[k1] * b[k2]
            poln.add_keyval(key, val)
    # c = c_n(n, n1, n2, n3, n4, n5, n6, n7, n8)
    c = comb(n, [n1, n2, n3, n4, n5, n6, n7, n8])
    return c * poln


def sub_v(poly):
    """Substitute :math:`v_{1,n-1}` and :math:`v_{2,n-1}` with their moments

    :param Poly poly: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h',
       'v_{1,n-1}','theta1','sigma_v1', 'v_{2,n-1}','theta2','sigma_v2').
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h', 'theta1','sigma_v1', 'theta2','sigma_v2').
    :rtype: Poly
    """
    poly_sum = Poly()
    kf = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2')
    poly_sum.set_keyfor(kf)
    #
    poln = Poly()
    for k in poly:
        pol1 = moment_v(k[3])  # [theta1, sigma_v1^2/k1]
        for key in pol1:
            i, j = key
            # t0 = list(k[0]); t0.insert(0, (1,0,j)); t0 = tuple(t0)
            t0 = t_mul_t0((1, 0, j), k[0])
            #
            knw = (t0, k[1], k[2], k[4] + i, k[5] + 2 * j, k[6], k[7], k[8])
            val = poly[k] * pol1[key]
            poln.add_keyval(knw, val)
    for k in poln:
        pol2 = moment_v(k[5])  # [theta2, sigma_v2^2/k2]
        for key in pol2:
            i, j = key
            # t0 = list(k[0]); t0.insert(0, (0,1,j)); t0 = tuple(t0)
            t0 = t_mul_t0((0, 1, j), k[0])
            #
            knw = (t0, k[1], k[2], k[3], k[4], k[6] + i, k[7] + 2 * j)
            val = poln[k] * pol2[key]
            poly_sum.add_keyval(knw, val)
    return poly_sum


def cmoment_y(l):
    """Central moment of :math:`y_n` with order :math:`l`

    :param int l: order of the central moment.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h', 'theta1','sigma_v1', 'theta2','sigma_v2').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h',
          'v_{1,n-1}', 'theta1', 'sigma_v1', 'v_{2,n-1}', 'theta2', 'sigma_v2']
    poly.set_keyfor(kf)
    #
    for n1 in range(l, -1, -1):
        for n2 in range(l - n1, -1, -1):
            for n3 in range(l - n1 - n2, -1, -1):
                for n4 in range(l - n1 - n2 - n3, -1, -1):
                    for n5 in range(l - n1 - n2 - n3 - n4, -1, -1):
                        for n6 in range(l - n1 - n2 - n3 - n4 - n5, -1, -1):
                            for n7 in range(l - n1 - n2 - n3 - n4 - n5 - n6, -1, -1):
                                n8 = l - n1 - n2 - n3 - n4 - n5 - n6 - n7
                                poly.merge(moment_comb(l, n1, n2, n3, n4, n5, n6, n7, n8))
    poly_sum = sub_v(poly)
    poly_sum.remove_zero()
    return poly_sum


##########
# scalar and (partial) derivative
##########

def dt0(t0, wrt):
    """Partial derivative of the tuple encoding k1 and k2

    :param tuple t0: tuple encoding k1 and k2 as
       '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}'.
    :param str wrt: with respect to.
    :return: list of tuple
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}', val).
    :rtype: list
    """
    deriv = []
    if wrt == 'k1':
        for i in range(len(t0)):
            key = list(t0)
            k = t0[i]
            if k[2] != 0 and k[0] != 0:
                K = (k[0], k[1], k[2] + 1)
                key[i] = K
                val = -k[2] * k[0]
                deriv.append((tuple(key), val))
    elif wrt == 'k2':
        for i in range(len(t0)):
            key = list(t0)
            k = t0[i]
            if k[2] != 0 and k[1] != 0:
                K = (k[0], k[1], k[2] + 1)
                key[i] = K
                val = -k[2] * k[1]
                deriv.append((tuple(key), val))
    else:
        msg = f"wrt must be either 'k1' or 'k2' while {wrt} is supplied!"
        raise ValueError(msg)
    return deriv


def dpoly(poly, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param Poly poly: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h', 'theta1','sigma_v1','theta2','sigma_v2').
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h', 'theta1','sigma_v1','theta2','sigma_v2').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'theta1', 'sigma_v1', 'theta2', 'sigma_v2')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. k1 or k2
    if wrt == 'mu':
        return (pold)
    elif wrt == 'k1' or wrt == 'k2':
        for k in poly:
            t0 = k[0]
            if len(t0) != 0:
                deriv = dt0(t0, wrt)
                if len(deriv) != 0:
                    for t in deriv:
                        t0_new, val_new = t
                        knw = list(k)
                        knw[0] = t0_new
                        val = val_new * poly[k]
                        pold.add_keyval(tuple(knw), val)
        for k in poly:
            if wrt == 'k1':
                n = k[1][0]
            else:
                n = k[1][1]
            if n != 0:
                knw = list(k)
                knw[2] += 1
                val = (-n) * poly[k]
                pold.add_keyval(tuple(knw), val)
    # partial derivative w.r.t. other parameters
    elif wrt in ['theta1', 'sigma_v1', 'theta2', 'sigma_v2']:
        if wrt == 'theta1': i = 3
        if wrt == 'sigma_v1': i = 4
        if wrt == 'theta2': i = 5
        if wrt == 'sigma_v2': i = 6
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                val = k[i] * poly[k]
                pold.add_keyval(tuple(knw), val)
    else:
        candidates = "'k1','k2','theta1','sigma_v1','theta2','sigma_v2'"
        raise ValueError(f"wrt must be one of {candidates}!")
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h', 'theta1','sigma_v1','theta2','sigma_v2').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k1 = par['k1']
    k2 = par['k2']
    h = par['h']
    theta1 = par['theta1']
    theta2 = par['theta2']
    sigma_v1 = par['sigma_v1']
    sigma_v2 = par['sigma_v2']
    #
    value = 0
    for k in poly:
        val = poly[k]
        t0 = k[0]
        for t in t0:
            val *= (t[0] * k1 + t[1] * k2) ** (-t[2])
        t1 = k[1]
        val *= math.exp(-(t1[0] * k1 + t1[1] * k2) * h)
        val *= h ** k[2]
        val *= theta1 ** k[3]
        val *= sigma_v1 ** k[4]
        val *= theta2 ** k[5]
        val *= sigma_v2 ** k[6]
        value += val
    return value


def cm(l, par):
    """Central moment in scalar

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :return: scalar of the central moment.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    value = poly2num(cmoment, par)
    return value


def dcm(l, par, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    pold = dpoly(cmoment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_2fsv.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    cmom = cmoment_y(n)
    print(f"cmoment_y({n}) = \n")
    pprint(cmom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{cmom.keyfor}")



"""
Conditional Moments (I)

Conditional moments of the SRJD model, given the initial variance.
"""
import math
from ajdmom import Poly
from ajdmom.mdl_srjd.cond_ie_iez_mom import moment_ie_iez

def moment_x(n):
    r"""conditional moment of :math:`\sigma IE_t + IEZ_t`

    Note x is defined as :math:`x_t \equiv e^{kt}(v_t - \theta) - (v_0 - \theta)`,
    which equals to :math:`\sigma IE_t + IEZ_t`.

    :param integer n: order of the conditional moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-theta', 'theta',
      'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n+1):
        bino = math.comb(n, i)
        poln = moment_ie_iez(i, n-i)
        for k, v in poln.items():
            key = list(k)
            key[4] += i
            key = tuple(key)
            val = bino * v
            poly.add_keyval(key, val)
    poly.remove_zero()
    return poly

def moment_v(n):
    """conditional moment of the state :math:`v_t`

    :param integer n: order of the conditional moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-theta', 'theta',
      'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n+1):
        bino1 = math.comb(n, i)
        for j in range(n-i+1):
            bino2 = math.comb(n - i, j)
            poln = moment_x(n - i - j)
            for k, v in poln.items():
                key = list(k)
                key[0] -= n - j
                key[2] += i
                key[3] += j
                key = tuple(key)
                val = bino1 * bino2 * v
                poly.add_keyval(key, val)
    poly.remove_zero()
    return poly

def poly2num(poly, par):
    v0, k, theta = par['v0'], par['k'], par['theta']
    sigma, lmbd, mu_v = par['sigma'], par['lmbd'], par['mu_v']
    h = par['h']
    f = 0
    for K, v in poly.items():
        val = math.exp(K[0] * k * h) / (k ** K[1])
        val *= ((v0 - theta) ** K[2]) * (theta ** K[3])
        val *= (sigma ** K[4]) * (lmbd ** K[5]) * (mu_v ** K[6])
        #
        f += val * v  # v is a Fraction, produce a floating-point number
    return f

def m_x(n, par):
    moment = moment_x(n)
    value = poly2num(moment, par)
    return value

def m(n, par):
    moment = moment_v(n)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint
    n = 3
    # print(f"\nmoment_x({n}) = "); pprint(moment_x(n))
    print(f"\nmoment_v({n}) = "); pprint(moment_v(n))
    #
    v0, h = 0.5, 1  # failed
    k, theta, sigma = 0.1, 0.5, 0.05
    lmbd, mu_v = 1, 0.3
    par = {'v0': v0, 'h': h, 'k': k, 'theta': theta,
           'sigma': sigma, 'mu_v': mu_v, 'lmbd': lmbd}
    print(f"\nm({n}) = {m(n, par)}")



"""
Conditional Central Moments of the SVVJ model, given initial variance
"""
import math
from fractions import Fraction

from ajdmom import Poly
from ajdmom.utils import comb, simplify_rho
from ajdmom.mdl_svcj.cond_ieii_ieziziz_mom import moment_ieii_ieziziz


def b_n(n1, n2, n3, n4, n5, n6, n7):
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho', 'sqrt(1-rho^2)',
          'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n2 + 1):
        for j in range(n6 + 1):
            bino = math.comb(n2, i) * math.comb(n6, j)
            sign = (-1) ** (i + n5 + n6 + j)
            c = sign * Fraction(bino, 2 ** (n1 + i + n4 + n5 + n6 + n7))
            key = (-(n1 + n4) - j, n7, n1 + i + n4 + n5 + 2*n6 + n7,
                   0, 0, 0, n1 + i, n2 - i, n3,
                   n6 + n7, n6 + n7)
            poly.add_keyval(key, c)
    return poly


def cmoment_y(n):
    """conditional central moment of the SVVJ model :math:`y_t|v_0`

    :param integer n: order of the conditional moment
    :return: poly with ``keyfor`` =  ('e^{kt}', 't', 'k^{-}',
      'mu', 'v0-theta', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v']
    if n < 0:
        raise ValueError('n must be non-negative')
    p0 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(0, 1)})
    p0.set_keyfor(kf)
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    # special case
    if n == 0:
        return p1
    elif n == 1:
        return p0
    # typical cases
    kf.insert(8, 'sqrt(1-rho^2)')
    poly = Poly()
    poly.set_keyfor(kf)
    for n1 in range(n + 1):
        for n2 in range(n - n1 + 1):
            for n3 in range(n - n1 - n2 + 1):
                for n4 in range(n - n1 - n2 - n3 + 1):
                    for n5 in range(n - n1 - n2 - n3 - n4 + 1):
                        for n6 in range(n - n1 - n2 - n3 - n4 - n5 + 1):
                            n7 = n - n1 - n2 - n3 - n4 - n5 - n6
                            c = comb(n, [n1, n2, n3, n4, n5, n6, n7])
                            pol1 = b_n(n1, n2, n3, n4, n5, n6, n7)
                            # kf = ['e^{kt}', 't', 'k^{-}',
                            #       'mu', 'v0-theta', 'theta', 'sigma', 'rho', 'sqrt(1-rho^2)',
                            #        'lmbd', 'mu_v']
                            pol2 = moment_ieii_ieziziz(n1, n2, n3, n4, n5, 0)
                            # kf = ['e^{kt}','t','k^{-}',
                            #       'v0-theta','theta','sigma',
                            #       'lmbd','mu_v','mu_s','sigma_s']
                            poln = Poly()
                            poln.set_keyfor(pol1.keyfor)
                            for k, v in pol2.items():
                                key = (k[0], k[1], k[2],
                                       0, k[3], k[4], k[5], 0, 0,
                                       k[6], k[7])
                                poln.add_keyval(key, v)
                            poly.merge(c * (pol1 * poln))
    return simplify_rho(poly, 8)


def poly2num(poly, par):
    # kf = ['e^{kt}', 't', 'k^{-}',
    #       'mu', 'v0-theta', 'theta', 'sigma', 'rho',
    #       'lmbd', 'mu_v']
    k = par['k']
    mu, v0, theta, sigma, rho = par['mu'], par['v0'], par['theta'], par['sigma'], par['rho']
    lmbd, mu_v = par['lmbd'], par['mu_v']
    h = par['h']
    #
    f = 0
    for K, V in poly.items():
        val = math.exp(K[0] * k * h) * (h ** K[1]) / (k ** K[2])
        val *= (mu ** K[3]) * ((v0 - theta) ** K[4]) * (theta ** K[5])
        val *= (sigma ** K[6]) * (rho ** K[7])
        val *= (lmbd ** K[8]) * (mu_v ** K[9])
        f += val * V
    return f


def cm_y(n, par):
    cmoment = cmoment_y(n)
    value = poly2num(cmoment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    n = 3
    poly = cmoment_y(n)
    print(f"cmoment_y({n}) = "); pprint(poly); print(f"which is a poly with keyfor = {poly.keyfor}")

    v0 = 0.007569
    k = 3.46
    theta = 0.008
    sigma = 0.14
    rho = -0.82
    r = 0.0319
    lmbd = 0.47
    mu_v = 0.05

    mu = r
    h = 1

    par = {'v0': v0, 'mu': mu, 'k': k, 'theta': theta, 'sigma': sigma, 'rho': rho,
           'lmbd': lmbd, 'mu_v': mu_v,
           'h': h}

    for n in range(1, 4+1):
        print(f"cmoment({n}) = {cm_y(n, par)}")



"""
Module for generating a trajectory of samples from Heston SV model by
Euler approximation

To facilitate the verification of the correctness of our codes
by comparison the population moments (given by our package) and
their sample counterparts.
Here I define function :py:func:`~ajdmom.mdl_1fsv.euler.r1FSV`.

In addition, I created :py:func:`~ajdmom.mdl_1fsv.euler.r1FSV_iid` for
generating i.i.d. samples.
"""
import math
import numpy as np


def r1FSV(v0, par, N, n_segment=10, h=1):
    """Generate a trajectory of samples from mdl_1fsv by Euler approximation.

    :param float v0: value of the initial variance :math:`v(0)`.
    :param dict par: parameters in a dict.
    :param int N: target length of samples to generate.
    :param int n_segment: number of segments each interval is split into.
    :param float h: time interval between each two consecutive samples.
    :return: a sequence of samples.
    :rtype: numpy 1-D array.
    """
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma_v']
    rho = par['rho']
    #
    dlt = h / n_segment
    std = math.sqrt(dlt)
    #
    y_series = np.empty(N)
    #
    for i in range(N):
        I = 0
        I2 = 0
        IV = 0
        for j in range(n_segment):
            eps = math.sqrt(v0) * np.random.normal(0, std)
            eps2 = math.sqrt(v0) * np.random.normal(0, std)
            v = v0 + k * (theta - v0) * dlt + sigma * eps
            if v < 0:
                v = 0  # eps = -(v0 + k*(theta-v0)*dlt)/sigma
            I += eps
            I2 += eps2
            IV += v0 * dlt
            v0 = v
        y = mu * h - IV / 2 + rho * I + math.sqrt(1 - rho ** 2) * I2
        y_series[i] = y
    return y_series


def r1FSV_iid(v0, par, N, n_segment=10, tau=1):
    """Generate iid samples from mdl_1fsv by Euler approximation.

    :param float v0: value of the initial variance :math:`v(0)`.
    :param dict par: parameters in a dict.
    :param int N: number of i.i.d samples to generate.
    :param int n_segment: number of segments the interval is split into.
    :param float tau: maturity time, or time length.
    :return: list of i.i.d. samples.
    :rtype: numpy 1-D array.
    """
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma_v']
    rho = par['rho']
    #
    dlt = tau / n_segment
    std = math.sqrt(dlt)
    #
    y = np.empty(N)
    #
    v0_original = v0
    for i in range(N):
        I = 0
        I2 = 0
        IV = 0
        v0 = v0_original
        for j in range(n_segment):
            eps = math.sqrt(v0) * np.random.normal(0, std)
            eps2 = math.sqrt(v0) * np.random.normal(0, std)
            v = v0 + k * (theta - v0) * dlt + sigma * eps
            if v < 0:
                v = 0  # eps = -(v0 + k*(theta-v0)*dlt)/sigma
            I += eps
            I2 += eps2
            IV += v0 * dlt
            v0 = v
        y[i] = mu * tau - IV / 2 + rho * I + math.sqrt(1 - rho ** 2) * I2
    return y


"""
Covariance for the Two-Factor SV with jumps
"""
import math

from ajdmom.poly import Poly
from ajdmom.cpp_mom import mcpp
from ajdmom.itos_mom import t_mul_t0
from ajdmom.mdl_2fsv.cov import moment_yy as m_yy
from ajdmom.mdl_2fsvj.mom import (
    moment_y,
    dpoly,
    poly2num
)


def moment_yy(l1, l2):
    """Co-Moment :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`

    :param int l1: :math:`l_1` in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: :math:`l_2` in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2', 'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'theta1', 'sigma_v1', 'theta2', 'sigma_v2', 'lambda', 'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    for i in range(l1, -1, -1):
        for j in range(l2, -1, -1):
            pol1 = m_yy(i, j)
            # keyfor = '(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
            #   'e^{-(n1*k1+n2*k2)h}','h','mu',
            #   'theta1','sigma_v1','theta2','sigma_v2')
            pol2 = mcpp(l1 - i)
            pol3 = mcpp(l2 - j)
            # keyfor = ('lambda*h','mu_j','sigma_j')
            poln = pol2 * pol3
            c = math.comb(l1, i) * math.comb(l2, j)
            for k1 in pol1:
                for k2 in poln:
                    key = (k1[0], k1[1], k1[2] + k2[0], k1[3], k1[4], k1[5], k1[6], k1[7],
                           k2[0], k2[1], k2[2])
                    val = c * pol1[k1] * poln[k2]
                    poly.add_keyval(key, val)
    poly.remove_zero()
    return poly


def cov_yy(l1, l2):
    """Covariance :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`

    :param int l1: :math:`l_1` in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :param int l2: :math:`l_2` in :math:`E[y_n^{l_1}y_{n+1}^{l_2}]`.
    :return: poly with attribute ``keyfor`` =
       ('(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
       'e^{-(n1*k1+n2*k2)h}','h','mu',
       'theta1','sigma_v1','theta2','sigma_v2', 'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['(n_1m*k1+n_2m*k2)^{-i_m},...,(n_11*k1+n_21*k2)^{-i_1}',
          'e^{-(n1*k1+n2*k2)h}', 'h', 'mu',
          'theta1', 'sigma_v1', 'theta2', 'sigma_v2', 'lambda', 'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    #
    pol1 = moment_y(l1)  # keyfor = kf
    pol2 = moment_y(l2)  # keyfor = kf
    for k1 in pol1:
        for k2 in pol2:
            t0 = k1[0]  # simply: t0 = k1[0] + k2[0]    # first special one
            for t in k2[0]: t0 = t_mul_t0(t, t0)  # carefully treat
            t1 = (k1[1][0] + k2[1][0], k1[1][1] + k2[1][1])  # second special one
            #
            key = (t0, t1) + tuple(k1[i] + k2[i] for i in range(2, 11))
            val = pol1[k1] * pol2[k2]
            poly.add_keyval(key, val)
    #
    cov = moment_yy(l1, l2) - poly  # moment_yy(l1,l2) with keyfor = kf
    cov.remove_zero()
    return cov


##########
# scalar and (partial) derivative
##########

def cov(l1, l2, par):
    """Covariance in scalar

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :return: scalar of the covariance.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    value = poly2num(covariance, par)
    return value


def dcov(l1, l2, par, wrt):
    """Partial derivative of covariance w.r.t. parameter wrt

    :param int l1: *l1* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param int l2: *l2* in :math:`cov(y_n^{l_1},y_{n+1}^{l_2})`.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    covariance = cov_yy(l1, l2)
    pold = dpoly(covariance, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_2fsvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l1 = 1 if len(args) == 0 else int(args[0])
    l2 = 1 if len(args) <= 1 else int(args[1])
    cov_l1_l2 = cov_yy(l1, l2)
    print(f"cov_yy({l1},{l2}) = \n")
    pprint(cov_l1_l2)
    print(f"\nwhich is a Poly with attribute keyfor =\n{cov_l1_l2.keyfor}")



import math
import numpy as np

from ajdmom.mdl_srjd.euler import rSRJD
from ajdmom.mdl_srjd.cond2_mom import m
from ajdmom.mdl_srjd.cond2_cmom import cm

rng = np.random.default_rng()


def comp_moments(orders, par, y_series):
    mnames = []
    tms = []
    sms = []
    for n in orders:
        tm = m(n, par)  # theoretial moment calculated by the package
        sm = np.mean(y_series ** n)  # sample moment
        mnames.append(f'moment({n})')
        tms.append(tm)
        sms.append(sm)
    return ((mnames, tms, sms))


def comp_cmoments(orders, par, y_series):
    mnames = []
    tms = []
    sms = []
    diff = y_series - np.mean(y_series)
    for n in orders:
        tm = cm(n, par)  # theoretial cmoment calculated by the package
        sm = np.mean(diff ** n)  # sample cmoment
        mnames.append(f'cmoment({n})')
        tms.append(tm)
        sms.append(sm)
    return ((mnames, tms, sms))


def print_comp(comp):
    txt = ''.ljust(10)
    txt += 'derived moment'.rjust(15) + 'sample moment'.rjust(15)
    txt += 'diff'.rjust(12) + 'diff(in %)'.rjust(12) + '\n'
    #
    for i in range(len(comp[0])):
        name = comp[0][i]
        tm = comp[1][i]
        sm = comp[2][i]
        diff = abs(tm - sm)
        if tm == 0:
            diff_percent = math.nan
        else:
            diff_percent = diff / abs(tm)
        #
        txt += '{:<10}{:>15.4f}{:>15.4f}{:>12.4f}{:>12.0%}\n'.format(name,
                                                                     tm, sm, diff, diff_percent)
    print(txt)


# h should be in par, since it is needed for m(), cm()
v0, h = 0.007569, 1
k, theta, sigma = 3.46, 0.008, 0.14
mu_v, lmbd = 0.05, 0.47
par = {'v0': v0, 'h': h, 'k': k, 'theta': theta, 'sigma': sigma, 'mu_v': mu_v, 'lmbd': lmbd}
#
for numJ in [0, 1, 2, 3]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
    jumpsize = rng.exponential(mu_v, numJ)
    jumptime = rng.uniform(0, h, numJ)  # unsorted
    jumpsize, jumptime = tuple(jumpsize), tuple(sorted(jumptime))  # set as tuples
    par['jumpsize'] = jumpsize
    par['jumptime'] = jumptime
    #
    x = rSRJD(v0, k, theta, sigma, h, jumpsize, jumptime, 100 * 1000)
    #
    orders = [1, 2, 3, 4, 5]
    mom_comp = comp_moments(orders, par, x)
    print_comp(mom_comp)
    #
    orders = [1, 2, 3, 4, 5]
    cmom_comp = comp_cmoments(orders, par, x)
    print_comp(cmom_comp)



"""
Conditional Moment of the SVVJ model, given initial variance
"""
import math
from fractions import Fraction

from ajdmom import Poly
from ajdmom.mdl_svvj.cond_cmom import cmoment_y, poly2num


def moment_y(n):
    """conditional moment of the SVVJ model :math:`y_t|v_0`

    :param integer n: order of the conditional moment
    :return: poly with ``keyfor`` =  ('e^{kt}', 't', 'k^{-}',
      'mu', 'v0-theta', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'v0-theta', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    if n < 0:
        raise ValueError('n must be non-negative')
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    # special case
    if n == 0:
        return p1
    # E[y|v0] = (mu - theta/2)*t - (v0 - theta)*(1-e^{-kt})/(2k)
    #           + lmbd * mu_v (1-e^{-kt})/(2k^2) - lmbd t mu_v / (2k)
    mean_y = Poly({
        (0, 1, 0,   1, 0, 0, 0, 0,   0, 0): Fraction(1, 1),
        (0, 1, 0,   0, 0, 1, 0, 0,   0, 0): Fraction(-1, 2),
        (0, 0, 1,   0, 1, 0, 0, 0,   0, 0): Fraction(-1, 2),
        (-1,0, 1,   0, 1, 0, 0, 0,   0, 0): Fraction(1, 2),
        (0, 0, 2,   0, 0, 0, 0, 0,   1, 1): Fraction(1, 2),
        (-1,0, 2,   0, 0, 0, 0, 0,   1, 1): Fraction(-1, 2),
        (0, 1, 1,   0, 0, 0, 0, 0,   1, 1): Fraction(-1, 2)
    })
    mean_y.set_keyfor(kf)
    if n == 1:
        return mean_y
    # typical cases
    # y_t = mean_y + \overline{y}_t
    for i in range(n + 1):
        bino = math.comb(n, i)
        pol1 = p1 if i == 0 else mean_y ** i
        poln = bino * (pol1 * cmoment_y(n - i))
        poly.merge(poln)
    return poly

def m_y(n, par):
    moment = moment_y(n)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    n = 3
    poly = moment_y(n)
    print(f"moment_y({n}) = "); pprint(poly); print(f"which is a poly with keyfor = {poly.keyfor}")

    v0 = 0.007569
    k = 3.46
    theta = 0.008
    sigma = 0.14
    rho = -0.82
    r = 0.0319
    lmbd = 0.47
    mu_v = 0.05

    mu = r
    h = 1

    par = {'v0': v0, 'mu': mu, 'k': k, 'theta': theta, 'sigma': sigma, 'rho': rho,
           'lmbd': lmbd, 'mu_v': mu_v,
           'h': h}

    for n in range(1, 4+1):
        print(f"moment({n}) = {m_y(n, par)}")



"""
Moments of the SRJD model

Unconditional moments of the SRJD model.
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_srjd.cmom import cmoment_v


def expand(poly):
    """expand the 'E[v]' in moment polys

    :param Poly poly: moment poly with ``keyfor`` =
      ('k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v')
    :return: poly with ``keyfor`` =
      ('k^{-}', 'theta', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['k^{-}', 'theta', 'sigma', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k, v in poly.items():
        # E[v] = theta + lmbd * mu_v / k
        n = k[1]
        if n == 0:
            poln.add_keyval(k, v)
            continue
        for i in range(n + 1):
            bino = math.comb(n, i)
            key = (k[0] + i, n - i, k[2], k[3] + i, k[4] + i)
            poln.add_keyval(key, bino * v)
    return poln

def moment_v(n, simplify = True):
    r"""moment of the state :math:`v_t`

    unconditional moment, noting that :math:`v_t = (v_t - E[v_t]) + E[v_t]`,
    the frist moment :math:`E[v] = \theta + \lambda\mu_v/k`

    :param int n: order of the unconditional moment
    :param bool simplify: whether to simplify the moment
    :return: if simplify = True, poly with ``keyfor`` =
      ('k^{-}', 'theta', 'sigma', 'lmbd', 'mu_v'),
      otherwise,
      ('k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError('n must be 0 or positive')
    kf = ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    if n == 0:
        poly.add_keyval((0, 0, 0, 0, 0), Fraction(1, 1))
        return poly if not simplify else expand(poly)
    if n == 1:
        poly.add_keyval((0, 1, 0, 0, 0), Fraction(1, 1))
        return poly if not simplify else expand(poly)
    for i in range(n + 1):
        bino = math.comb(n, i)
        poln = cmoment_v(i)
        for k, v in poln.items():
            key = list(k)
            key[1] += n - i  # (E[v_t])^{n-i}
            key = tuple(key)
            poly.add_keyval(key, bino * v)
    return poly if not simplify else expand(poly)


if __name__ == "__main__":
    from pprint import pprint

    print('\nExample usage of the module functions\n')
    n = 4
    mom = moment_v(n, simplify = False)
    print(f'moment_v({n}) = ')
    pprint(mom)
    print(f"{mom.keyfor}")



r"""
Conditional Central Moment of :math:`E[IE_t^{n_1} \overline{IEZ}_t^{n_2}|v_0]`
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_srjd.iez_cmom import cmoment_IEZ


def iez_to_ie_iez(poly):
    # from
    # kf = ['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v']
    # to (s = 0)
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        key = (k[0], k[2], 0, 0, 0, k[3], k[4])
        poln.add_keyval(key, v)
    return poln

def int_e_poly(c, tp, m, poly):
    poln = Poly()
    poln.set_keyfor(['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v'])
    for k, v in poly.items():
        n = k[0] + m  # \int_0^t e^{n ks} ds = e^{n kt} /(nk) - 1/(nk)
        if n == 0:
            raise ValueError(r"n = 0, the integral should be \int_0^t ds = t!")
        key = list(k)
        key[0] = n
        key[1] += 1
        if tp in [1, 2, 3]: key[tp + 1] += 1
        key = tuple(key)
        poln.add_keyval(key, c * v * Fraction(1, n))
        #
        key = list(k)
        key[0] = 0
        key[1] += 1
        if tp in [1, 2, 3]: key[tp + 1] += 1
        key = tuple(key)
        poln.add_keyval(key, c * v * Fraction(-1, n))
    return poln

def recursive_ie(n, IE, IE_IEZ):
    poly = Poly()
    poly.set_keyfor(['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v'])
    if n >= 2:
        c = Fraction(n * (n - 1), 2)
        poly.merge(int_e_poly(c, 1, 1, IE[n - 2]))
        poly.merge(int_e_poly(c, 2, 2, IE[n - 2]))
        poly.merge(int_e_poly(c, 3, 1, IE[n - 1]))
        poly.merge(int_e_poly(c, 4, 1, IE_IEZ[(n - 2, 1)]))
    return poly

def moment_ie(n):
    """Moment of :math:`E[IE_t^n]`

    :param integer n: order of the moment

    :return: poly with ``keyfor`` =
      ('e^{kt}', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError(f"moment_ie({n}) is called!")
    #
    # IE: a dict of moments of E[IE^n]
    #
    IE = {}
    #
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    # special poly constants, analog to 0 and 1
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, 0): Fraction(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    P1.set_keyfor(kf)
    #
    # n = 0: special case
    #
    IE[0] = P1  # equiv to constant 1
    #
    # n = 1
    #
    IE[1] = P0  # equiv to constant 0
    #
    P2 = Poly({
        (1, 1, 1, 0, 0, 0, 0): Fraction(1, 1),
        (0, 1, 1, 0, 0, 0, 0): Fraction(-1, 1),
        (2, 1, 0, 1, 0, 0, 0): Fraction(1, 2),
        (0, 1, 0, 1, 0, 0, 0): Fraction(-1, 2)
    })
    IE[2] = P2
    #
    if n <= 2:
        return IE[n]
    #
    IE_IEZ = {}
    #
    IE_IEZ[(0, 1)] = iez_to_ie_iez(cmoment_IEZ(1))
    #
    IE_IEZ[(1, 1)] = P0
    #
    # n >= 3: typical cases
    #
    if n > 3:
        # compute all lower-order moments to get ready
        for i in range(3, n):
            if i > 3: IE_IEZ[(i - 2, 1)] = cmoment_ie_iez(i - 2, 1)
            poly = recursive_ie(i, IE, IE_IEZ)
            poly.remove_zero()
            IE[i] = poly
            # delete polys no more needed
            del IE[i - 2]
            del IE_IEZ[(i - 2, 1)]
    # the last one
    if n > 3: IE_IEZ[(n - 2, 1)] = cmoment_ie_iez(n - 2, 1)
    poly = recursive_ie(n , IE, IE_IEZ)
    poly.remove_zero()
    return poly

def key_times_poly(k, poly):
    # k: ['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v']
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for key, val in poly.items():
        knw = (k[0] + key[0], k[2] + key[1], key[2], key[3], key[4],
               k[3] + key[5], k[4] + key[6])
        poln.add_keyval(knw, val)
    return poln

def recursive_ie_iez(n1, n2, IE_IEZ):
    poly = Poly()
    poly.set_keyfor(['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v'])
    if n1 >= 2 and n2 >= 1:
        for i in range(n2 + 1):
            bino = math.comb(n2, i)
            pol1 = cmoment_IEZ(n2 - i)  # ['e^{kt}', 'e^{ks}', 'k^{-}', 'lmbd', 'mu_v']
            for k, v in pol1.items():
                coef = bino * Fraction(n1 * (n1 - 1), 2) * v
                pol2  = int_e_poly(coef, 1, 1 + k[1], IE_IEZ[(n1 - 2, i)])
                pol2 += int_e_poly(coef, 2, 2 + k[1], IE_IEZ[(n1 - 2, i)])
                pol2 += int_e_poly(coef, 3, 1 + k[1], IE_IEZ[(n1 - 1, i)])
                pol2 += int_e_poly(coef, 4, 1 + k[1], IE_IEZ[(n1 - 2, i + 1)])
                poly.merge(key_times_poly(k, pol2))
    if n1 >= 2 and n2 == 0:
        poly = moment_ie(n1)
    if n1 == 1:  # no matter what value n2 take, the result is 0
        key, val = (0, 0, 0, 0, 0, 0, 0), Fraction(0, 1)
        poly.add_keyval(key, val)
    if n1 == 0:  # works even n2 = 0
        poly = iez_to_ie_iez(cmoment_IEZ(n2))
    return poly

def cmoment_ie_iez(n1, n2):
    r"""Conditional Central Moment of :math:`E[IE_t^{n_1} \overline{IEZ}_t^{n_2}|v_0]`

    :param integer n1: order of :math:`IE_t`
    :param integer n2: order of :math:`\overline{IEZ}_t`
    :return: poly with ``keyfor`` =
        ('e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if not (n1 >= 0 and n2 >= 0):
        raise ValueError(f'moment_ie_iez({n1}, {n2}) is called!')
    #
    # IE_IEZ: a dict of moments of E[IE^{n1} IEZ^{n2}]
    #
    IE_IEZ = {}
    #
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    # special poly constants, analog to 0 and 1
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, 0): Fraction(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    P1.set_keyfor(kf)
    #
    # n1 + n2 = 0: special case
    #
    IE_IEZ[(0, 0)] = P1  # equiv to constant 1
    #
    # n1 + n2 = 1
    #
    IE_IEZ[(1, 0)] = P0  # equiv to constant 0
    IE_IEZ[(0, 1)] = iez_to_ie_iez(cmoment_IEZ(1))
    #
    # n1 + n2 = 2
    #
    IE_IEZ[(2, 0)] = moment_ie(2)
    IE_IEZ[(1, 1)] = P0
    IE_IEZ[(0, 2)] = iez_to_ie_iez(cmoment_IEZ(2))
    #
    if n1 + n2 <= 2:
        return IE_IEZ[(n1, n2)]
    #
    # n1 + n2 >= 3: typical cases
    #
    if n1 + n2 > 3:
        # compute all lower-order moments to get ready
        for n in range(3, n1 + n2):
            for i in range(n + 1):
                poly = recursive_ie_iez(i, n - i, IE_IEZ)
                poly.remove_zero()
                IE_IEZ[(i, n - i)] = poly
            # seems almost all polys should be reserved
    # the last one
    poly = recursive_ie_iez(n1, n2, IE_IEZ)
    poly.remove_zero()
    return poly

def poly2num(poly, par):
    # 'e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v'
    v0, k, theta, sigma = par['v0'], par['k'], par['theta'], par['sigma']
    lmbd, mu_v, h = par['lmbd'], par['mu_v'], par['h']
    #
    meanv = theta + lmbd * mu_v / k
    #
    f = 0
    for K, V in poly.items():
        val = math.exp(K[0] * k * h) / (k ** K[1])
        val *= ((v0-meanv) ** K[2]) * (meanv ** K[3]) * (sigma ** K[4])
        val *= (lmbd ** K[5]) * (mu_v ** K[6])
        f += val * V
    return f

def m_ie(n, par):
    moment = moment_ie(n)
    value = poly2num(moment, par)
    return value

def cm_ie_iez(n1, n2, par):
    moment = cmoment_ie_iez(n1, n2)
    value = poly2num(moment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint

    # pprint(iez_to_ie_iez(moment_IEZ(n)))

    # n = 6
    # print(f"moment_ie(n) = ")
    # moment_ie(10)
    # pprint(moment_ie(n))

    print('\nExample usage of the module functions')
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    n1 = 3
    n2 = 2
    print(f"\ncmoment_ie_iez(n1, n2) returns a poly with keyfor = \n{kf}\n")
    print(f'cmoment_ie_iez({n1}, {n2}) = ')
    pprint(cmoment_ie_iez(n1, n2))



"""
Central Moments of the SRJD model

Unconditional central moments of the SRJD model.
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_srjd.cond_ie_iez_cmom import cmoment_ie_iez

def cond_cmoment_v(n):
    """Conditional central moment :math:`E[(v_t - E[v])^n|v0]`

    Different from the cmoment_v() in 'cond_cmom.py' in the ``keyfor``
    attribute.

    :param int n: order of the conditional central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-E[v]',
      'E[v]', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    if n == 0:
        return p1
    if n == 1:
        key = (-1, 0, 1, 0, 0, 0, 0)
        poly.add_keyval(key, Fraction(1, 1))
        return poly
    for i1 in range(n + 1):
        for i2 in range(n - i1 + 1):
            i3 = n - i1 - i2
            bino = math.comb(n, i1) * math.comb(n - i1, i2)
            poln = cmoment_ie_iez(i1, i2)
            for k, v in poln.items():
                key = list(k)
                key[0] -= n   # e^{-nkt}
                key[2] += i3  # v0-E[v]
                key[4] += i1  # sigma
                key = tuple(key)
                poly.add_keyval(key, v * bino)
    return poly

def find_same_order(poly, n):
    """Check the poly does have one and only one monomial with order n

    :param Poly poly: conditional central moment of the state with
      ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma',
      'lmbd', 'mu_v')
    :param integer n: order of :math:`v0-E[v]`
    :return: True or raise an error
    :rtype: bool
    """
    count = 0
    target_key = (-n, 0, n, 0, 0, 0, 0)
    target_order = n
    for k, v in poly.items():
        if k[2] == target_order:
            count += 1
    if count == 1 and poly[target_key] ==  Fraction(1, 1):
        return True
    else:
        msg = "Either multiple of monomials (v0-E[v])^n, or coefficient not correct"
        raise Exception(msg)

def key_times_poly(coef, k, poly):
    # k   0        1          2        3        4        5       6
    # ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #                     to expand
    #
    # poly: expanded E[(v0-E[v])^i] with keyfor =
    # ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    kf = ['e^{kt}', 'k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    for K, V in poly.items():
        key = (k[0],
               k[1] + K[0],
               k[3] + K[1],
               k[4] + K[2],
               k[5] + K[3],
               k[6] + K[4])
        poln.add_keyval(key, coef * V)
    return poln

def last_expectation(poly, uncond_cm):
    # poly with keyfor =
    # ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #                    expectation
    #
    kf = ['e^{kt}', 'k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    poln = Poly()
    poln.set_keyfor(kf)
    #
    for k, v in poly.items():
        order = k[2]
        if order == 0:   # constant
            key = k[0:2] + k[3:]
            poln.add_keyval(key, v)
        # order == 1, with expectation = 0
        elif order > 1:
            poln.merge(key_times_poly(v, k, uncond_cm[order]))
    return poln

def divide(n, poly):
    # poly with keyfor =
    # ['e^{kt}', 'k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    kf = ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    pol1 = Poly()
    pol1.set_keyfor(kf)
    pol2 = Poly()
    pol2.set_keyfor(kf)
    #
    for k, v in poly.items():
        if v == 0: continue  # works for Fraction(0, 1)
        #
        if k[0] == 0:
            pol1.add_keyval(k[1:], v)
        elif k[0] == -n:
            pol2.add_keyval(k[1:], v)
        else:
            raise ValueError("k[0] is neither 0 nor -n")
    #
    if len(pol1) != len(pol2):
        raise ValueError("len(pol1) != len(pol2)")
    #
    for k in pol1:
        if pol1[k] != -pol2[k]:
            raise ValueError("pol1[k] != -pol2[k]")
    # double check
    for k in pol2:
        if pol2[k] != -pol1[k]:
            raise ValueError("pol2[k] != -pol1[k]")
    #
    return pol1

def solve_polynomial(n, uncond_cm):
    #
    # get the conditional central moment as a polynomial of v0-E[v]
    #
    cond_cm = cond_cmoment_v(n)
    # kf = ['e^{kt}', 'k^{-}', 'v0-E[v]', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    #
    # pick up the special monomial: e^{-nkt} (v0-E[v])^n
    #
    if find_same_order(cond_cm, n):
        del cond_cm[(-n, 0, n, 0, 0, 0, 0)]
    else:
        raise Exception("error with the conditional central moment")
    #
    # get the unconditional central moment in the left hand of the equation
    #
    uncond_left = last_expectation(cond_cm, uncond_cm)
    #
    # divide the left by 1 - e^{-nkt}
    #
    poly = divide(n, uncond_left) # ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    return poly

def cmoment_v(n):
    """central moment :math:`E[(v_t - E[v])^m]`

    unconditional central moment

    :param int n: order of the unconditional central moment
    :return: poly with ``keyfor`` = ('k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError("n must be >= 0")
    kf = ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    p1 = Poly({(0 ,0 ,0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    p0 = Poly({(0 ,0 ,0, 0, 0): Fraction(0, 1)})
    p0.set_keyfor(kf)
    #
    uncond_cm = []
    uncond_cm.append(p1)  # E[(v_0 - E[v])^0]
    uncond_cm.append(p0)  # E[(v_0 - E[v])^1]
    #
    if n < 2:
        return uncond_cm[n]
    #
    for i in range(2, n):
        poly = solve_polynomial(i, uncond_cm)
        uncond_cm.append(poly)
    poly = solve_polynomial(n, uncond_cm)
    return poly


if __name__ == "__main__":
    from pprint import pprint

    print('\nExample usage of the module functions\n')
    kf = ['k^{-}', 'E[v]', 'sigma', 'lmbd', 'mu_v']
    n = 4
    print(f'cmoment_v({n}) = ')
    pprint(cmoment_v(n))
    print(f"{kf}")



"""
Conditional Central Moments (I)

Conditional central moments of the SRJD model, given the initial variance.
"""
import math
from ajdmom import Poly
from ajdmom.mdl_srjd.cond_mom import moment_x, poly2num

def cmoment_x(n):
    r"""conditional central moment of :math:`\sigma IE_t + IEZ_t`

    Note x is defined as :math:`x_t \equiv e^{kt}(v_t - \theta) - (v_0 - \theta)`,
    which equals to :math:`\sigma IE_t + IEZ_t`.

    :param integer n: order of the conditional central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-theta', 'theta',
      'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n+1):
        bino1 = math.comb(n, i)
        poln = moment_x(i)
        for j in range(n - i + 1):
            bino2 = math.comb(n - i, j)
            for k, v in poln.items():
                key = list(k)
                key[0] += j
                key[1] += n - i
                key[5] += n - i
                key[6] += n - i
                key = tuple(key)
                val = bino1 * bino2 * v * ((-1) ** j)
                poly.add_keyval(key, val)
    poly.remove_zero()
    return poly

def cmoment_v(n):
    """conditional central moment of the state :math:`v_t`

    :param integer n: order of the conditional central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 'k^{-}', 'v0-theta', 'theta',
      'sigma', 'lmbd', 'mu_v')
    :rtype: Poly
    """
    kf = ['e^{kt}', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'lmbd', 'mu_v']
    poly = Poly()
    poly.set_keyfor(kf)
    for i in range(n+1):
        bino1 = math.comb(n, i)
        for j in range(n-i+1):
            bino2 = math.comb(n - i, j)
            poln = moment_x(i)
            for k, v in poln.items():
                key = list(k)
                key[0] -= i + j
                key[1] += n - i
                key[5] += n - i
                key[6] += n - i
                key = tuple(key)
                val = bino1 * bino2 * ((-1) ** (n-i-j)) * v
                poly.add_keyval(key, val)
    poly.remove_zero()
    return poly

def cm_x(n, par):
    cmoment = cmoment_x(n)
    value = poly2num(cmoment, par)
    return value

def cm(n, par):
    cmoment = cmoment_v(n)
    value = poly2num(cmoment, par)
    return value


if __name__ == '__main__':
    from pprint import pprint
    n = 4
    # print(f"\ncmoment_x({n}) = "); pprint(cmoment_x(n))
    print(f"\ncmoment_v({n}) = "); pprint(cmoment_v(n))
    #
    v0, h = 0.5, 1  # failed
    k, theta, sigma = 0.1, 0.5, 0.05
    lmbd, mu_v = 1, 0.3
    par = {'v0': v0, 'h': h, 'k': k, 'theta': theta,
           'sigma': sigma, 'mu_v': mu_v, 'lmbd': lmbd}
    print(f"\ncm({n}) = {cm(n, par)}")



"""
Central Moments for One-Factor SV
"""
# from pprint import pprint
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.utils import comb
from ajdmom.ito_mom import moment_v, moment_IEII


def b_n(n1, n2, n3, n4, n5):
    r"""Represent :math:`b(\boldsymbol{n})` in :eq:`b-n` in Poly

    :param int n1: times of :math:`\theta` being selected.
    :param int n2: times of :math:`v_{n-1}` being selected.
    :param int n3: times of :math:`eI_n` being selected.
    :param int n4: times of :math:`I_n` being selected.
    :param int n5: times of :math:`I_n^{*}` being selected.
    :return: a poly with attribute ``keyfor`` =
       ('e^{-kh}','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    b = Poly()
    b.set_keyfor(['e^{-kh}', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)'])
    for i in range(n1 + n2 + 1):
        for j in range(n4 + 1):
            key = (i, n1 + n2 + n3 + j, n1, n3 + j, n4 - j, n5)
            num = math.comb(n1 + n2, i) * math.comb(n4, j) * ((-1) ** (n2 + i + j))
            den = 2 ** (n1 + n2 + n3 + j)
            b.add_keyval(key, Frac(num, den))
    return b


def ve_IEII(n2, n3, n4, n5):
    r""":math:`v_{n-1}^{n_2} e^{-n_3knh} IEII`

    :param int n2: :math:`n_2` in above formula.
    :param int n3: :math:`n_3` in :math:`IEII` which is
       :math:`E[eI_n^{n_3}I_n^{n_4}I_n^{*n_5}|v_{n-1}]`.
    :param int n4: :math:`n_4` in :math:`IEII`.
    :param int n5: :math:`n_5` in :math:`IEII`.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v').
    :rtype: Poly
    """
    poln = Poly()
    poln.set_keyfor(['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'theta', 'sigma_v'])
    #
    poly = moment_IEII(n3, n4, n5)
    # with keyfor
    # ['e^{k(n-1)h}','e^{kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    # and the power for the first one always = n3
    for k in poly.keys():
        # × e^{-n_3knh}      × v_{n-1}^{n_2}
        key = (n3 - k[1], k[2], n2 + k[3], k[4], k[5], k[6])
        poln.add_keyval(key, poly[k])
    return poln


def moment_comb(n, n1, n2, n3, n4, n5):
    r"""Moment for this combination in expansion of :math:`\overline{y}_n^l`

    :param int n: l in :math:`E[\overline{y}_{n}^l]`.
    :param int n1: times of :math:`\theta` being selected.
    :param int n2: times of :math:`v_{n-1}` being selected.
    :param int n3: times of :math:`eI_n` being selected.
    :param int n4: times of :math:`I_n` being selected.
    :param int n5: times of :math:`I_n^{*}` being selected.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = ve_IEII(n2, n3, n4, n5)
    # ['e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v']
    b = b_n(n1, n2, n3, n4, n5)
    # ['e^{-kh}','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)']
    keyfor = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'theta', 'sigma_v', 'rho',
              'sqrt(1-rho^2)']  # keyfor for new poly
    keyIndexes = ([0, 1, 2, 3, 4, 5, -1, -1], [0, -1, -1, 1, 2, 3, 4, 5])
    # -1 for not having this term
    poly = poly.mul_poly(b, keyIndexes, keyfor)
    c = comb(n, [n1, n2, n3, n4, n5])
    return c * poly


def sub_v(poly):
    r"""Substitute :math:`v_{n-1}` with its moment (poly) in Central Moment of y_n

    :param Poly poly: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','v_{n-1}','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)')
    :rtype: Poly
    """
    poly_sum = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)']
    poly_sum.set_keyfor(kf)
    for k in poly:
        poln = moment_v(k[2])  # ['theta','sigma_v^2/k']
        for key in poln:
            i, j = key
            knw = (k[0], k[1], k[3] + j, k[4] + i, k[5] + 2 * j, k[6], k[7])
            val = poln[key] * poly[k]
            poly_sum.add_keyval(knw, val)
    return poly_sum


def cmoment_y(l):
    r"""Central moment of :math:`y_n` of order :math:`l`

    :param int l: order of the central moment, l >= 0.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'v_{n-1}', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)']
    poly.set_keyfor(kf)
    for n1 in range(l, -1, -1):
        for n2 in range(l - n1, -1, -1):
            for n3 in range(l - n1 - n2, -1, -1):
                for n4 in range(l - n1 - n2 - n3, -1, -1):
                    n5 = l - n1 - n2 - n3 - n4
                    poly.merge(moment_comb(l, n1, n2, n3, n4, n5))
    poly_sum = sub_v(poly)
    poly_sum.remove_zero()
    return poly_sum


##########
# scalar and (partial) derivative
##########
def dpoly(poly, wrt):
    r"""Partial derivative of central moment w.r.t. parameter wrt

    :param Poly poly: central moment represented by the poly.
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('e^{-kh}', 'h', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. mu
    if wrt == 'mu':
        return (pold)
    # partial derivative w.r.t. k
    elif wrt == 'k':
        for k in poly:
            if k[0] != 0:
                knw = list(k)
                knw[1] += 1
                knw = tuple(knw)
                val = (-k[0]) * poly[k]
                pold.add_keyval(knw, val)
            if k[2] != 0:
                knw = list(k)
                knw[2] += 1
                knw = tuple(knw)
                val = (-k[2]) * poly[k]
                pold.add_keyval(knw, val)
    elif wrt in ['theta', 'sigma_v']:
        if wrt == 'theta': i = 3
        if wrt == 'sigma_v': i = 4
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                knw = tuple(knw)
                val = k[i] * poly[k]
                pold.add_keyval(knw, val)
    # partial derivative w.r.t. rho
    elif wrt == 'rho':
        for k in poly:
            if k[5] != 0:
                knw = list(k)
                knw[5] -= 1
                knw = tuple(knw)
                val = k[5] * poly[k]
                pold.add_keyval(knw, val)
            if k[6] != 0:
                knw = list(k)
                knw[5] += 1
                knw[6] -= 2
                knw = tuple(knw)
                val = (-k[6]) * poly[k]
                pold.add_keyval(knw, val)
    else:
        raise ValueError("wrt must be either 'k', 'theta', 'sigma_v' or 'rho'!")
    return pold


def poly2num(poly, par):
    r"""Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    # ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)')
    k = par['k']
    h = par['h']
    theta = par['theta']
    sigma_v = par['sigma_v']
    rho = par['rho']
    #
    value = 0
    for K in poly:
        val = poly[K] * math.exp(-K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (theta ** K[3]) * (sigma_v ** K[4]) * (rho ** K[5])
        val *= (1 - rho ** 2) ** (K[6] / 2)
        value += val
    return value


def cm(l, par):
    """Central moment in scalar

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :return: scalar of the central moment.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    value = poly2num(cmoment, par)
    return value


def dcm(l, par, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    pold = dpoly(cmoment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsv.py' for more test
    from pprint import pprint
    from ajdmom.utils import simplify_rho
    import sys

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    cmom = cmoment_y(n)
    # cmom = simplify_rho(cmom, 6)
    # cmom.write_to_txt(f'heston-cmoment-{n}-formula-simplified.txt')
    print(f"cmoment_y({n}) = \n")
    pprint(cmom)
    print(f"\nwhich is a Poly with attribute keyfor = \n{cmom.keyfor}")
    #
    par = {'mu': 0.125, 'k': 0.1, 'theta': 0.25, 'sigma_v': 0.1, 'rho': -0.7, 'h': 1}
    val = poly2num(cmom, par)
    print(f"\nnumerical value = {val}, given parameters = \n{par}")



## Formula Deriviation - Moment
from ajdmom import mdl_1fsv
from pprint import pprint

m1 = mdl_1fsv.moment_y(1)
msg = f"which is a Poly with attribute keyfor = \n{m1.keyfor}"
print("moment_y(1) = ")
pprint(m1)
print(msg)



"""
Central Moments for One-Factor SV with jumps
"""
import math

from ajdmom.poly import Poly
from ajdmom.mdl_1fsv.cmom import cmoment_y as cm_y
# ['e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)']
from ajdmom.cpp_mom import cmcpp


# ['lambda*h','mu_j','sigma_j^2']

def cmoment_y(l):
    """Central moment of :math:`y_n` of order :math:`l`

    :param int l: order of the moment.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)','lambda',
       'mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)', 'lambda',
          'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    # if l == 0:
    #   poln = Poly({(0,0,0,0,0,0,0,0,0,0): 1}); poln.set_keyfor(kf); return(poln)
    #
    for i in range(l + 1):
        coef = math.comb(l, i)
        pol1 = cm_y(i)
        # ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)')
        pol2 = cmcpp(l - i)
        # ('lambda*h','mu','sigma')
        keyIndexes = [(0, 1, 2, 3, 4, 5, 6, -1, -1, -1), (-1, 0, -1, -1, -1, -1, -1, 0, 1, 2)]
        poln = pol1.mul_poly(pol2, keyIndexes, kf)
        poly.merge(coef * poln)
    return poly


##########
# scalar and (partial) derivative
##########

def dpoly(poly, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param Poly poly: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)','lambda',
       'mu_j','sigma_j')
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)','lambda',
       'mu_j','sigma_j').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('e^{-kh}', 'h', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)', 'lambda',
          'mu_j', 'sigma_j')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. mu
    if wrt == 'mu':
        return pold
    # partial derivative w.r.t. k
    elif wrt == 'k':
        for k in poly:
            if k[0] != 0:
                knw = list(k)
                knw[1] += 1
                knw = tuple(knw)
                val = (-k[0]) * poly[k]
                pold.add_keyval(knw, val)
            if k[2] != 0:
                knw = list(k)
                knw[2] += 1
                knw = tuple(knw)
                val = (-k[2]) * poly[k]
                pold.add_keyval(knw, val)
    # partial derivative w.r.t. theta or sigma_v
    elif wrt in ['theta', 'sigma_v', 'lambda', 'mu_j', 'sigma_j']:
        if wrt == 'theta': i = 3
        if wrt == 'sigma_v': i = 4
        if wrt == 'lambda': i = 7
        if wrt == 'mu_j': i = 8
        if wrt == 'sigma_j': i = 9
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                knw = tuple(knw)
                val = k[i] * poly[k]
                pold.add_keyval(knw, val)
    # partial derivative w.r.t. rho
    elif wrt == 'rho':
        for k in poly:
            if k[5] != 0:
                knw = list(k)
                knw[5] -= 1
                knw = tuple(knw)
                val = k[5] * poly[k]
                pold.add_keyval(knw, val)
            if k[6] != 0:
                knw = list(k)
                knw[5] += 1
                knw[6] -= 2
                knw = tuple(knw)
                val = (-k[6]) * poly[k]
                pold.add_keyval(knw, val)
    else:
        candidates = "'k','theta','sigma_v','rho','lambda','mu_j','sigma_j'"
        raise ValueError(f"wrt must be one of {candidates}!")
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','theta','sigma_v','rho','sqrt(1-rho^2)','lambda',
       'mu_j','sigma_j')
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k = par['k']
    h = par['h']
    theta = par['theta']
    sigma_v = par['sigma_v']
    rho = par['rho']
    lmbd = par['lambda']
    mu_j = par['mu_j']
    sigma_j = par['sigma_j']
    #
    value = 0
    for K in poly:
        val = poly[K] * math.exp(-K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (theta ** K[3]) * (sigma_v ** K[4]) * (rho ** K[5])
        val *= (1 - rho ** 2) ** (K[6] / 2)
        val *= (lmbd ** K[7]) * (mu_j ** K[8]) * (sigma_j ** K[9])
        value += val
    return value


def cm(l, par):
    """Central moment in scalar

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :return: scalar of the central moment.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    value = poly2num(cmoment, par)
    return value


def dcm(l, par, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param int l: order of the central moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    cmoment = cmoment_y(l)
    pold = dpoly(cmoment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    #
    kf = ('e^{-kh}', 'h', 'k^{-}', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)',
          'lambda', 'mu_j', 'sigma_j^2')
    print(f"cmoment_y(l) returns a poly with keyfor = \n{kf}")
    args = sys.argv[1:]
    n = 2 if len(args) == 0 else int(args[0])
    print(f"cmoment_y({n}) = ")
    pprint(cmoment_y(n))



import numpy as np
import math

from ajdmom.mdl_svcj.euler import rSVCJ
from ajdmom.mdl_svcj.cond2_mom import m
from experiment3 import comp_moments, print_comp

rng = np.random.default_rng()


v0, h = 0.007569, 1
k, theta, sigma, rho = 3.46, 0.008, 0.14, -0.82
mu = 0.0319
lmbd_v, mu_v = 0.47, 0.05
mu_b = -0.1
lmbd_s, sigma_s, rhoJ = lmbd_v, 0.0001, -0.38
#
mu_s = math.log((1 + mu_b) * (1 - rhoJ * mu_v)) - sigma_s ** 2 / 2
# same as that in SVJ in Broadie-Kaya (2006)
par = {'v0': v0, 'h': h, 'mu': mu, 'k': k, 'theta': theta,
       'sigma': sigma, 'rho': rho, 'lmbd_v': lmbd_v, 'mu_v': mu_v,
       'lmbd_s': lmbd_s, 'mu_s': mu_s, 'sigma_s': sigma_s, 'rhoJ': rhoJ}

for numJ in [2]:  # numJ = rng.poisson(lmbd * h): the actual number of jumps
    jsize_v = rng.exponential(mu_v, numJ)
    jtime_v = rng.uniform(0, h, numJ)  # unsorted
    jsize_v, jtime_v = tuple(jsize_v), tuple(sorted(jtime_v))  # set as tuples
    par['jumpsize'] = jsize_v
    par['jumptime'] = jtime_v
    #
    print(f"\nWith conditions v0 = {v0}, \njumpsize = {jsize_v}, \njumptime = {jtime_v}")
    #
    n, n_segment = 4000 * 1000, 100
    y = rSVCJ(v0, mu, k, theta, sigma, rho, h, jsize_v, jtime_v,
              mu_s, sigma_s, rhoJ, n, n_segment)
    orders = [1, 2, 3, 4, 5]
    mom_comp = comp_moments(orders, par, y, m)
    print(f"------------------------------moment_ie------------------------------")
    print_comp(mom_comp)



import math
import numpy as np

from ajdmom.mdl_svcj.cond_mom import m_y
from ajdmom.mdl_svcj.euler import r_y


def comp_moments(orders, par, x, m):
    x = np.array(x)
    mnames, tms, sms = [], [], []
    for n in orders:
        tm = m(n, par)  # theoretical moment calculated by the package
        sm = np.mean(x ** n)  # sample moment
        mnames.append(f'moment({n})')
        tms.append(tm)
        sms.append(sm)
    return mnames, tms, sms


def print_comp(comp):
    txt = ''.ljust(10)
    txt += 'derived moment'.rjust(15) + 'sample moment'.rjust(15)
    txt += 'diff'.rjust(12) + 'diff(in %)'.rjust(12) + '\n'
    for i in range(len(comp[0])):
        name = comp[0][i]
        tm = comp[1][i]
        sm = comp[2][i]
        diff = abs(tm - sm)
        diff_percent = math.nan if tm == 0 else diff / abs(tm)
        temp = '{:<10}{:>15.7f}{:>15.7f}{:>12.7f}{:>12.0%}\n'
        txt += temp.format(name, tm, sm, diff, diff_percent)
    print(txt[0:-1])


h = 1
v0, k, theta, sigma, rho = 0.007569, 3.46, 0.008, 0.14, -0.82
lmbd, mu_b, sigma_s, mu_v, rhoJ = 0.47, -0.1, 0.0001, 0.05, -0.38
r = 0.0319

mu_s = math.log((1+mu_b)*(1-rhoJ*mu_v)) - sigma_s ** 2 / 2
mu = r - lmbd * mu_b

N = 4000 * 1000

par = {'v0': v0, 'mu': mu, 'k': k, 'theta': theta, 'sigma': sigma, 'rho': rho,
       'lmbd': lmbd, 'mu_v': mu_v, 'rhoJ': rhoJ, 'mu_s': mu_s, 'sigma_s': sigma_s,
       'h': h}

print(f'mu = {mu}, mu_s = {mu_s}')

nsegment = 100
y_series = r_y(v0, mu, k, theta, sigma, rho, lmbd, mu_v, rhoJ, mu_s, sigma_s, h, N, nsegment)
orders = [1, 2, 3, 4, 5]
mom_comp = comp_moments(orders, par, y_series, m_y)
print(f"------------------------------moment_ie------------------------------")
print_comp(mom_comp)



"""
Central Moments of the SVCJ model
"""
import math
from fractions import Fraction

from ajdmom.poly import Poly
from ajdmom.mdl_svcj.mom import moment_y, poly2num

def cmoment_y(n):
    """central moment of :math:`y_t`

    Central moment :math:`E[\bar{y}_t^n] = E[(y_t - E[y_t])^n]`

    :param int n: order of the central moment
    :return: poly with ``keyfor`` = ('e^{kt}', 't', 'k^{-}',
      'mu', 'E[v]', 'theta', 'sigma', 'rho',
      'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s')
    :rtype: Poly
    """
    if n < 0:
        raise ValueError('n must be 0 or positive')
    mean_y = moment_y(1)
    kf = ['e^{kt}', 't', 'k^{-}',
          'mu', 'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    p0 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(0, 1)})
    p0.set_keyfor(kf)
    p1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0): Fraction(1, 1)})
    p1.set_keyfor(kf)
    #
    if n == 0:
        return p1
    if n == 1:
        return p0
    # typical cases
    for i in range(n+1):
        bino = math.comb(n, i)
        sign = (-1) ** i
        pol1 = p1 if i == 0 else mean_y ** i
        poln = (bino * sign) * (pol1 * moment_y(n - i))
        poly.merge(poln)
    return poly

def cm_y(n, par):
    cmoment = cmoment_y(n)
    value = poly2num(cmoment, par)
    return value

def remove_mu_column(poly):
    poly.remove_zero()
    kf = ['e^{kt}', 't', 'k^{-}',
          #'mu',
          'E[v]', 'theta', 'sigma', 'rho',
          'lmbd', 'mu_v', 'rhoJ', 'mu_s', 'sigma_s']
    kf[0] = 'e^{-kt}'
    poln = Poly()
    poln.set_keyfor(kf)
    for k, v in poly.items():
        if k[3] == 0:
            key = (-k[0], k[1], k[2]) + tuple(k[4:])
            poln.add_keyval(key, v)
        else:
            raise Exception(f'k = {k} where k[3] != 0, i.e., mu exist')
    return poln


if __name__ == '__main__':
    from pprint import pprint

    n = 4
    cmom = cmoment_y(n)
    cmom = remove_mu_column(cmom)
    print(f"cmoment_y({n}) =")
    pprint(cmom)
    print(f"{len(cmom)} rows of {cmom.keyfor}")



"""
Conditional Moments for the SVCJ model, given :math:`v_0`
and the realized jumps in the variance.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.cpp_mom import mnorm
from ajdmom.utils import comb, fZ
from ajdmom.mdl_svvj.cond2_mom import moments_y_to as moments_y_to_svvj


def mnorm_cond(n):
    r"""conditional normal distribution moment :math:`J_i^s|J_i^v \sim
    \mathcal{N} (\mu_s+rhoJ J_i^v, \sigma_s^2)`

    :param int n: order of the moment, n >= 0
    :return: poly with ``keyfor`` = ('mu_s', 'rhoJ', 'sigma_s', 'J_i')
    :rtype: Poly
    """
    kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_i']
    # special case
    if n == 0:
        P1 = Poly({(0, 0, 0, 0): Frac(1, 1)})
        P1.set_keyfor(kf)
        return (P1)
    # n >=1
    poly = Poly()
    poly.set_keyfor(kf)
    #
    poln = mnorm(n)  # kf = ['mu','sigma^2']
    for k in poln:
        for i in range(k[0] + 1):
            c = math.comb(k[0], i)
            key = (i, k[0] - i, 2 * k[1], k[0] - i)
            val = c * poln[k]
            poly.add_keyval(key, val)
    return poly


def comb_poly(index):
    r"""multiply together those conditional normal distribution moments

    :param list index: a combination of the Compound Poisson Process power :math:`m`,
       :math:`(m_1,\dots,m_{N(t)})`
    :return: poly with attribute ``keyfor`` =
      ('mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}')
    :rtype: Poly
    """
    kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    poly.add_keyval((0, 0, 0, ()), Frac(1, 1))
    for i in index:
        # mnorm_cond(0) supported
        pol1 = mnorm_cond(i)  # ['mu_s', 'rhoJ', 'sigma_s', 'J_i']
        poln = Poly()
        poln.set_keyfor(kf)
        for k1 in pol1:
            for k2 in poly:
                key = (k1[0] + k2[0], k1[1] + k2[1], k1[2] + k2[2], k2[3] + (k1[3],))
                val = pol1[k1] * poly[k2]
                poln.add_keyval(key, val)
        poly = poln
    return poly


def moment_IZs(m, J):
    r"""moment :math:`\mathbb{E}[(IZ_t^{s})^{m}|z^v(u), 0\le u \le t]`

    :param int m: power of :math:`IZ^s_t`
    :param tuple J: jump sizes in the variance
    :return: poly with attribute ``keyfor`` =
      ('mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}')
    :rtype: Poly
    """
    kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}']
    poly = Poly()
    poly.set_keyfor(kf)
    #
    numJ = len(J)
    #
    # special cases
    #
    if m == 0:  # return(1)
        poly.add_keyval((0, 0, 0, tuple(0 for i in range(numJ))), Frac(1, 1))
        return (poly)
    if numJ == 0:  # return(0)
        poly.add_keyval((0, 0, 0, ()), Frac(0, 1))
        return (poly)
    #
    # typical cases
    #
    if numJ == 1:  # single normal variable
        poln = mnorm_cond(m)  # kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_i']
        for k in poln:
            key = (k[0], k[1], k[2], (k[3],))
            poly.add_keyval(key, poln[k])
        return (poly)
    if numJ == 2:
        for i in range(0, m + 1):
            c = math.comb(m, i)
            pol1 = mnorm_cond(i)
            pol2 = mnorm_cond(m - i)  # mnorm_cond(0) supported
            # kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_i']
            for k1 in pol1:
                for k2 in pol2:
                    key = (k1[0] + k2[0], k1[1] + k2[1], k1[2] + k2[2], (k1[3], k2[3]))
                    val = c * pol1[k1] * pol2[k2]
                    poly.add_keyval(key, val)
        return (poly)
    # numJ > 2
    stack = []
    nextstep = "forward"
    while len(stack) < numJ:
        if nextstep == "forward":
            if len(stack) < numJ - 2:
                stack.append(0)
                nextstep = "forward"
            else:
                left = m - sum(stack)
                for i in range(left + 1):
                    stack.extend([i, left - i])
                    index = stack.copy()
                    #
                    c = comb(m, index)
                    poln = comb_poly(index)
                    poly += c * poln
                    #
                    stack.pop()
                    stack.pop()
                # checking whether all indexes traversed
                if stack[0] == m: break
                #
                nextstep = "backward"
        else:
            left = m - sum(stack)
            if left == 0:
                stack.pop()
                nextstep = "backward"
            else:
                stack[-1] += 1
                nextstep = "forward"
    return poly


def moments_y_to(l, J):
    r"""conditional moments of :math:`y_t` of orders :math:`0:l`

    :param int l: highest order of the conditional moments to derive, l >= 1
    :param tuple J: jump sizes in the variance
    :return: a list of polys with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}'), noting that 'p' encodes power
      of :math:`I\!Z_t`.
    :rtype: list
    """
    moms = []
    moms_svvj = moments_y_to_svvj(l)  # 0-th to l-th moments
    #
    kf = moms_svvj[0].keyfor + ('mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}')
    #
    # special case: 0-th moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, 0, 0, (), (), 0, 0, 0, 0, 0, 0, 0, ()): Frac(1, 1)})
    P1.set_keyfor(kf)
    moms.append(P1)  # equiv to constant 1
    #
    # typical cases
    #
    for n in range(1, l + 1):
        poly = Poly()
        poly.set_keyfor(kf)
        for n1 in range(n, -1, -1):
            c = math.comb(n, n1)
            pol1 = moms_svvj[n1]  # 0-th moment supported
            pol2 = moment_IZs(n - n1, J)  # kf = ['mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}']
            for k1 in pol1:
                for k2 in pol2:
                    key = k1 + k2
                    poly.add_keyval(key, c * pol1[k1] * pol2[k2])
        poly.remove_zero()
        moms.append(poly)
    return moms


def moment_y(l, J):
    moms = moments_y_to(l, J)
    return moms[-1]


##########
# scalar
##########

def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','beta_t','mu-theta/2','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'p',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)',
      'mu_s', 'rhoJ', 'sigma_s', 'J_{1:n}')
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    mu = par['mu']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    rho = par['rho']
    h = par['h']  # t= h
    J = par['jumpsize']  # vector of jump sizes
    s = par['jumptime']  # vector of jump time points
    mu_s = par['mu_s']
    rhoJ = par['rhoJ']
    sigma_s = par['sigma_s']
    #
    beta_t = (1 - math.exp(-k * h)) / (2 * k)
    #
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (beta_t ** K[3]) * ((mu - theta / 2) ** K[4]) * ((v0 - theta) ** K[5])
        val *= (theta ** K[6]) * (sigma ** K[7])
        #
        l, o, p= K[8], K[9], K[10]
        #
        val *= fZ(l, o, k, s, J)
        #
        IZ = sum(J)
        val *= IZ ** p
        #
        val *= ((sigma / (2 * k)) ** K[11]) * ((rho - sigma / (2 * k)) ** K[12])
        val *= ((1 - rho ** 2) ** (K[13] / 2))
        #
        val *= (mu_s ** K[14]) * (rhoJ ** K[15]) * (sigma_s ** K[16])
        #
        f_J = 1
        for i in range(len(J)):
            f_J *= (J[i]) ** (K[17][i])
        val *= f_J
        #
        f += val * poly[K]
    return f


def m(l, par):
    """conditional moment in scalar

    :param int l: order of the conditional moment
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the conditional central moment
    :rtype: float
    """
    J = par['jumpsize']  # list of jump sizes
    moments = moments_y_to(l, J)
    moment = moments[-1]
    value = poly2num(moment, par)
    return value


if __name__ == "__main__":
    from pprint import pprint
    import sys
    import random

    print('Example usage of the module function\n')
    args = sys.argv[1:]
    numJ = 2 if len(args) == 0 else int(args[0])  # special cases: 0, 1, 2
    jumpsize = [random.expovariate(lambd=1.0) for i in range(numJ)]
    print(f"jumpsize = ")
    pprint(jumpsize)
    #
    l = 2 if len(args) <= 1 else int(args[1])
    moms = moments_y_to(l, jumpsize)  # 0-th to l-th moments
    kf = moms[0].keyfor
    print(f"\nmoment_y({l}, J) = \n")
    pprint(moms[l])
    print(f"\nwhich is a poly with keyfor = \n{kf}")



import pytest
import math
from ajdmom.poly import Poly

from ajdmom.utils import comb, dbfactorial, cmnorm, simplify_rho


def test_comb():
    expected = math.comb(5, 1) * math.comb(5 - 1, 2)
    actual = comb(5, [1, 2, 2])
    assert actual == expected


def test_dbfactorial():
    n = [0, 1, 2, 3, 10, 11]
    expected = [1, 1, 2, 3, 3840, 10395]
    for i in range(len(n)):
        assert dbfactorial(n[i]) == expected[i]


def test_cmnorm():
    coef = [1, 0, 1, 0, 3, 0, 15, 0, 105]
    power = [0, 0, 2, 0, 4, 0, 6, 0, 8]
    for i in range(len(coef)):
        cmom = cmnorm(i)
        for key in cmom:
            pow_actual, coef_actual = key[0], cmom[key]
            assert coef_actual == coef[i]
            assert pow_actual == power[i]


def test_simplify_rho():
    kf = ['rho', 'sqrt(1-rho^2)']
    poly = Poly({
        (1, 0): 1,
        (0, 2): 1
    })
    poly.set_keyfor(kf)
    #
    poly_expected = Poly({
        (0,): 1,
        (1,): 1,
        (2,): -1
    })
    poly_expected.set_keyfor(kf)
    #
    poly_actual = simplify_rho(poly, 1)
    assert poly_actual == poly_expected



"""
Conditional Central Moments for the SVVJ model, given the initial state of
the variance and the realized jumps in the variance.

Conditional central moments are derived simultaneously because the one-by-one
procedure is not efficient.
"""
import math
from fractions import Fraction as Frac

from ajdmom.poly import Poly
from ajdmom.utils import simplify_rho, comb, fZ
from ajdmom.ito_cond2_mom import recursive_IEII


def simplify(poly, tp=0):
    """Simplify the final poly

    :param Poly poly: poly with ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :param int tp: type of the simplification:
      (1) ``tp = 0``: only changes from 'e^{kt}' to 'e^{-kt}'.
      (2) ``tp = 1``: changes in addition from 'sigma/2k', 'rho-sigma/2k',
      'sqrt(1-rho^2)' to 'rho'.
    :return: poly with ``keyfor`` =
      ('e^{-kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}', 'rho') when ``tp = 1``;
      poly with ``keyfor`` changes only from 'e^{kt}' to 'e^{-kt}'
      if ``tp = 0``.
    :rtype: Poly
    """
    keyfor = ('e^{-kt}',) + poly.keyfor[1:]
    poln = Poly()
    #
    # tp = 0: only changes from e^{kt} to e^{-kt}
    #
    if tp == 0:
        poln.set_keyfor(keyfor)
        for k, v in poly.items():
            poln.add_keyval((-k[0],) + k[1:], v)
        return poln
    #
    # tp = 1: simplify more
    #   from 'sigma/2k' (8), 'rho-sigma/2k' (9) to 'rho'
    kf = keyfor[0:8] + ('rho',) + keyfor[10:]
    poln.set_keyfor(kf)
    for k in poly:
        for i in range(0, k[9] + 1):
            p = k[8] + i  # power of 'sigma/2k'
            key = (-k[0], k[1], k[2] + p, k[3], k[4], k[5] + p,
                   k[6], k[7], k[9] - i, k[10])
            val = math.comb(k[9], i) * ((-1) ** i) * Frac(1, 2 ** p) * poly[k]
            poln.add_keyval(key, val)
    #   from 'rho' (9), 'sqrt(1-rho^2)' (10) to 'rho'
    poln = simplify_rho(poln, 10)
    return poln


def cmoments_y_to(l, show=False):
    """Derive the central moments with orders from 0 to l simultaneously

    :param int l: highest order of the conditional central moments to derive, l >= 1
    :param bool show: show the in-process message or not, defaults to False
    :return: a list of the conditional central moments with ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :rtype: list
    """
    # IEII: a dict of moments of E[IE_t^{n1}I_t^{n2}I_t^{*n3}]
    IEII = {}
    kf = ['e^{kt}', 't', 'k^{-}', 'v0-theta', 'theta', 'sigma', 'l_{1:n}', 'o_{1:n}']
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, (), ()): Frac(0, 1)})
    P0.set_keyfor(kf)
    P1 = Poly({(0, 0, 0, 0, 0, 0, (), ()): Frac(1, 1)})
    P1.set_keyfor(kf)
    #
    # n1 + n2 + n3 = 0
    IEII[(0, 0, 0)] = P1  # support for special case, constant 1
    # n1 + n2 + n3 = 1
    IEII[(1, 0, 0)] = P0  # support for special case, constant 0
    IEII[(0, 1, 0)] = P0
    IEII[(0, 0, 1)] = P0
    # n1 + n2 +n3 = 2
    P200 = Poly({
        (1, 0, 1, 1, 0, 0, (), ()): Frac(1, 1),
        (0, 0, 1, 1, 0, 0, (), ()): -Frac(1, 1),
        (2, 0, 1, 0, 1, 0, (), ()): Frac(1, 2),
        (0, 0, 1, 0, 1, 0, (), ()): -Frac(1, 2),
        (1, 0, 1, 0, 0, 0, (0,), (0,)): Frac(1, 1),
        (0, 0, 1, 0, 0, 0, (1,), (0,)): -Frac(1, 1)
    })
    P200.set_keyfor(kf)
    P110 = Poly({
        (0, 1, 0, 1, 0, 0, (), ()): Frac(1, 1),
        (1, 0, 1, 0, 1, 0, (), ()): Frac(1, 1),
        (0, 0, 1, 0, 1, 0, (), ()): -Frac(1, 1),
        (0, 1, 0, 0, 0, 0, (0,), (0,)): Frac(1, 1),
        (0, 0, 0, 0, 0, 0, (0,), (1,)): -Frac(1, 1)
    })
    P110.set_keyfor(kf)
    P020 = Poly({
        (-1, 0, 1, 1, 0, 0, (), ()): -Frac(1, 1),
        (0, 0, 1, 1, 0, 0, (), ()): Frac(1, 1),
        (0, 1, 0, 0, 1, 0, (), ()): Frac(1, 1),
        (-1, 0, 1, 0, 0, 0, (0,), (0,)): -Frac(1, 1),
        (0, 0, 1, 0, 0, 0, (-1,), (0,)): Frac(1, 1)
    })
    P020.set_keyfor(kf)
    #
    IEII[(2, 0, 0)] = P200
    IEII[(1, 1, 0)] = P110
    IEII[(1, 0, 1)] = P0
    IEII[(0, 2, 0)] = P020
    IEII[(0, 1, 1)] = P0
    IEII[(0, 0, 2)] = P020
    #
    kf += ['sigma/2k', 'rho-sigma/2k', 'sqrt(1-rho^2)']
    #
    cmoms = []
    #
    # special case: 0-th central moment
    #
    P1 = Poly({(0, 0, 0, 0, 0, 0, (), (), 0, 0, 0): Frac(1, 1)})
    P1.set_keyfor(kf)
    cmoms.append(P1)  # equiv to constant 1
    #
    # special case: 1-th central moment
    #
    P0 = Poly({(0, 0, 0, 0, 0, 0, (), (), 0, 0, 0): Frac(0, 1)})
    P0.set_keyfor(kf)
    cmoms.append(P0)  # equiv to constant 0
    #
    # typical cases
    #
    for n in range(2, l + 1):
        # container for the n-th conditional central moment of y
        poly = Poly()
        poly.set_keyfor(kf)
        # iter over all combinations of i + j + (n-i-j) = n
        for i in range(n, -1, -1):
            for j in range(n - i, -1, -1):
                # derive for this exact combination
                c = comb(n, [i, j, n - i - j])  # constant
                if n < 3:
                    poln = IEII[(i, j, n - i - j)]  # already in IEII
                else:
                    poln = recursive_IEII(i, j, n - i - j, IEII)
                # add this exact combination into poly
                # c * b * poln
                for k in poln:
                    key = list(k)
                    key[0] -= i
                    key = tuple(key)  # × e^{-ikt}
                    key = key + (i, j, n - i - j)  # enlarge with last 3
                    poly.add_keyval(key, c * poln[k])
                # reserve poln for further use or delete it otherwise
                if n >= 3 and n < l:
                    IEII[(i, j, n - i - j)] = poln
                else:
                    del poln
        # delete polys when no more needed
        if n == l:
            del IEII
        else:  # only reserve the most recent polys
            index = [key for key in IEII if key[0] + key[1] + key[2] == n - 2]
            for key in index: del IEII[key]
        poly.remove_zero()
        cmoms.append(poly)
        if show: print(f"complete the derivation for the {n}-th central moment.")
    return cmoms


def cmoment_y(l):
    """Derive the conditional central moment

    :param int l: order of the conditional central moment to derive
    :return: Poly object with ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :rtype: Poly
    """
    cmoments = cmoments_y_to(l)
    return cmoments[-1]


def classify(poly):
    """Classify the key-value pairs inside poly according to levels of summation

    :param Poly poly: poly with the first part of its ``keyfor`` =
     ('e^{-kt}','t','k^{-}','v0-theta','theta','sigma',
     'l_{1:n}', 'o_{1:n}')
     and the second part may be ('sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
     or ('rho')
    :return: list of subpolys
    :rtype: list
    """
    # record levels of summations
    N_sum = []
    for k in poly:
        if len(k[6]) not in N_sum: N_sum.append(len(k[6]))
    N_sum = sorted(N_sum)  # 0,1,2,...
    #
    # excludes l_{1:n}, o_{1:n}
    kf0 = poly.keyfor[:6] + poly.keyfor[8:]
    subpolys = [Poly() for n in N_sum]
    #
    subpolys[0].set_keyfor(kf0 if N_sum[0] == 0 else poly.keyfor)
    # exclude the sub-keyfor if the sub-key corresponding to 'l_{1:n}' is empty
    for i in range(1, len(N_sum)): subpolys[i].set_keyfor(poly.keyfor)
    #
    # classify
    for k, v in poly.items():
        subpoly = subpolys[len(k[6])]
        subpoly[k[0:6] + k[8:] if len(k[6]) == 0 else k] = v
        # exclude the empty subkey (),()
    return subpolys


def write_to_subpolys(poly, nth):
    """Write poly into different subpolys according to levels of summation

    :param Poly poly: poly with ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :param int nth: order of the moment
    :return: nothing
    :rtype: None
    """
    # only changes from e^{kt} to e^{-kt}
    poly = simplify(poly, tp=0)  # do not over simplify
    subpolys = classify(poly)
    for n in range(len(subpolys)):
        subpoly = subpolys[n]
        fname = f"svvj-cond-cmoment-{nth}-formula-{n}.txt"
        subpoly.write_to_txt(fname)


##########
# scalar
##########
def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
      ('e^{kt}','t','k^{-}','v0-theta','theta','sigma',
      'l_{1:n}', 'o_{1:n}',
      'sigma/2k','rho-sigma/2k','sqrt(1-rho^2)')
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the poly
    :rtype: float
    """
    v0 = par['v0']
    k = par['k']
    theta = par['theta']
    sigma = par['sigma']
    rho = par['rho']
    h = par['h']  # t = h
    J = par['jumpsize']  # vector of jump sizes
    s = par['jumptime']  # vector of jump time points
    #
    f = 0
    for K in poly:
        val = math.exp(K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= ((v0 - theta) ** K[3]) * (theta ** K[4]) * (sigma ** K[5])
        #
        l, o = K[6], K[7]
        val *= fZ(l, o, k, s, J)
        #
        val *= (sigma / (2 * k)) ** K[8]
        val *= (rho - sigma / (2 * k)) ** K[9]
        val *= (1 - rho ** 2) ** (K[10] / 2)
        #
        f += val * poly[K]
    return f


def cm(l, par):
    """conditional central moment in scalar

    :param int l: order of the conditional central moment
    :param dict par: parameters in dict, ``jumptime`` (tuple) and ``jumpsize`` (tuple)
      must also be included
    :return: scalar of the central moment
    :rtype: float
    """
    cmoments = cmoments_y_to(l)
    cmoment = cmoments[-1]  # need only the last one
    value = poly2num(cmoment, par)
    return value


if __name__ == "__main__":
    import sys
    from pprint import pprint

    print('\nExample usage of the module function\n')
    args = sys.argv[1:]
    l = 3 if len(args) == 0 else int(args[0])
    cmoms = cmoments_y_to(l, show=True)  # 0-th to l-th central moments
    kf = cmoms[0].keyfor
    print(f"\ncmoments_y_to(l) returns a list of polys with keyfor = \n{kf}")
    # for i in range(1, l+1):
    #   cmom = cmoms[i]
    #   write_to_subpolys(cmom, i)
    print(f"\ncmoment_y({l}) = ")
    pprint(cmoms[l])
    print(f"which is a poly with keyfor = \n{kf}")



"""
Package ajdmom

This package is devoted to compute the theoretical moments and covariances of
some Affine Jump Diffusion processes, such as Heston Stochastic Volatility
models and its derived models, such as including jumps in the return process
and so on.
"""
from .poly import Poly
from .cpp_mom import mcpp, cmcpp
from .ito_mom import moment_v, moment_IEII
from .itos_mom import moment_IEI_IEII
# from .ito_cond2_mom import moment_IEII



"""
Module for generating samples from :abbr:`SVIJ(Stochastic Volatility with
Independent Jumps in the price and variance)`

Note that we are generating i.i.d. samples with fixed conditions instead of
a trajectory of samples!
"""
import math
import numpy as np
from ajdmom.mdl_svvj.euler import rSVVJ

rng = np.random.default_rng()


def rSVIJ(v0, mu, k, theta, sigma, rho, T, jsize_v, jtime_v,
          lmbd_s, mu_s, sigma_s, n, n_seqment):
    r"""Generate i.i.d. Random samples from mdl_svij by Euler approximation

    :param float v0: value of the initial variance :math:`v(0)`
    :param float mu: parameter :math:`\mu`
    :param float k: parameter :math:`k`
    :param float theta: parameter :math:`\theta`
    :param float sigma: parameter :math:`\sigma_v`
    :param float rho: parameter :math:`\rho`
    :param float T: end time
    :param tuple jsize_v: jump sizes in the variance over [0, T]
    :param tuple jtime_v: jump time points in the variance over [0, T]
    :param float lmbd_s: jump arrival rate in the price
    :param float mu_s: mean of the normal distribution of jumps in the price
    :param float sigma_s: standard deviation of the normal distribution of jumps
      in the price
    :param int n: number of samples to generate
    :param int n_seqment: number of segments to split [0, T]
    :return: i.i.d. samples
    :rtype: numpy.ndarray
    """
    y_svvj = rSVVJ(v0, mu, k, theta, sigma, rho, T, jsize_v, jtime_v, n, n_seqment)
    #
    IZ = np.empty(n)
    for i in range(n):
        numJ = rng.poisson(lmbd_s * T)
        jsize_y = rng.normal(mu_s, sigma_s, numJ)
        IZ[i] = sum(jsize_y)
    return y_svvj + IZ


if __name__ == "__main__":
    import sys

    v0, h = 0.007569, 1
    r, k, theta, sigma, rho = 0.0319, 3.46, 0.008, 0.14, -0.82
    lmbd_v, mu_v = 0.47, 0.05  # same as that in mdl_svcj
    #
    mu_b = -0.12
    lmbd_s, sigma_s = 0.11, 0.15
    mu_s = math.log(1 + mu_b) - sigma_s ** 2 / 2
    # same as that in SVJ in Broadie-Kaya (2006)
    #
    args = sys.argv[1:]
    numJ = rng.poisson(lmbd_v * h) if len(args) == 0 else int(args[0])
    jsize_v = rng.exponential(mu_v, numJ)
    jtime_v = rng.uniform(0, h, numJ)
    jtime_v = sorted(jtime_v)
    # change to tuples
    jsize_v, jtime_v = tuple(jsize_v), tuple(jtime_v)
    #
    n, n_segment = 20, 10
    y = rSVIJ(v0, r, k, theta, sigma, rho, h, jsize_v, jtime_v,
              lmbd_s, mu_s, sigma_s, n, n_segment)
    print(f"\nGiven v0 = {v0} and {numJ} jumps in the variance, ")
    print(f"jsize_v = {jsize_v}, \njtime_v = {jtime_v}")
    print(f"\n{n} i.i.d. samples from SVIJ: ")
    print(y)



"""
Moments for One-Factor SV with jumps
"""
import math

from ajdmom.poly import Poly
from ajdmom.mdl_1fsv.mom import moment_y as m_y
# ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)')
from ajdmom.cpp_mom import mcpp


# ('lambda*h','mu','sigma^2')

def moment_y(l):
    """Moment of :math:`y_n` of order :math:`l`

    :param int l: order of the moment.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    poly = Poly()
    kf = ['e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)',
          'lambda', 'mu_j', 'sigma_j']
    poly.set_keyfor(kf)
    #
    for i in range(l + 1):
        coef = math.comb(l, i)
        pol1 = m_y(i)
        # ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)')
        pol2 = mcpp(l - i)
        # ('lambda*h','mu','sigma')
        keyIndexes = [(0, 1, 2, 3, 4, 5, 6, 7, -1, -1, -1), (-1, 0, -1, -1, -1, -1, -1, -1, 0, 1, 2)]
        poln = pol1.mul_poly(pol2, keyIndexes, kf)
        poly.merge(coef * poln)
    return poly


##########
# scalar and (partial) derivative
##########

def dpoly(poly, wrt):
    """Partial derivative of central moment w.r.t. parameter wrt

    :param Poly poly: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j')
    :param str wrt: with respect to.
    :return: poly with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j').
    :rtype: Poly
    """
    pold = Poly()
    kf = ('e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)',
          'lambda', 'mu_j', 'sigma_j')
    pold.set_keyfor(kf)
    #
    # partial derivative w.r.t. k
    if wrt == 'k':
        for k in poly:
            if k[0] != 0:
                knw = list(k)
                knw[1] += 1
                val = (-k[0]) * poly[k]
                pold.add_keyval(tuple(knw), val)
            if k[2] != 0:
                knw = list(k)
                knw[2] += 1
                val = (-k[2]) * poly[k]
                pold.add_keyval(tuple(knw), val)
    # partial derivative w.r.t. rho
    elif wrt == 'rho':
        for k in poly:
            if k[6] != 0:
                knw = list(k)
                knw[6] -= 1
                val = k[6] * poly[k]
                pold.add_keyval(tuple(knw), val)
            if k[7] != 0:
                knw = list(k)
                knw[6] += 1
                knw[7] -= 2
                val = (-k[7]) * poly[k]
                pold.add_keyval(tuple(knw), val)
    elif wrt in ['mu', 'theta', 'sigma_v', 'lambda', 'mu_j', 'sigma_j']:
        if wrt == 'mu': i = 3
        if wrt == 'theta': i = 4
        if wrt == 'sigma_v': i = 5
        if wrt == 'lambda': i = 8
        if wrt == 'mu_j': i = 9
        if wrt == 'sigma_j': i = 10
        for k in poly:
            if k[i] != 0:
                knw = list(k)
                knw[i] -= 1
                val = k[i] * poly[k]
                pold.add_keyval(tuple(knw), val)
    else:
        candidates = "'k','mu','theta','sigma_v','rho','lambda','mu_j','sigma_j'"
        raise ValueError(f"wrt must be one of {candidates}.")
    return pold


def poly2num(poly, par):
    """Decode poly back to scalar

    :param Poly poly: poly to be decoded with attribute ``keyfor`` =
       ('e^{-kh}','h','k^{-}','mu','theta','sigma_v','rho','sqrt(1-rho^2)',
       'lambda','mu_j','sigma_j').
    :param dict par: parameters in dict.
    :return: scalar of the poly.
    :rtype: float
    """
    k = par['k']
    h = par['h']
    mu = par['mu']
    theta = par['theta']
    sigma_v = par['sigma_v']
    rho = par['rho']
    lmbd = par['lambda']
    mu_j = par['mu_j']
    sigma_j = par['sigma_j']
    #
    value = 0
    for K in poly:
        val = poly[K] * math.exp(-K[0] * k * h) * (h ** K[1]) * (k ** (-K[2]))
        val *= (mu ** K[3])
        val *= (theta ** K[4]) * (sigma_v ** K[5]) * (rho ** K[6])
        val *= (1 - rho ** 2) ** (K[7] / 2)
        val *= (lmbd ** K[8]) * (mu_j ** K[9]) * (sigma_j ** K[10])
        value += val
    return value


def m(l, par):
    """Moment in scalar

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :return: scalar of the moment.
    :rtype: float
    """
    moment = moment_y(l)
    value = poly2num(moment, par)
    return value


def dm(l, par, wrt):
    """Partial derivative of moment w.r.t. parameter wrt

    :param int l: order of the moment.
    :param dict par: parameters in dict.
    :param str wrt: with respect to.
    :return: scalar of the partial derivative.
    :rtype: float
    """
    moment = moment_y(l)
    pold = dpoly(moment, wrt)
    value = poly2num(pold, par)
    return value


if __name__ == "__main__":
    # Example usage of the module, see 'tests/test_mdl_1fsvj.py' for more test
    from pprint import pprint
    import sys

    print('\nExample usage of the module function\n')
    kf = ('e^{-kh}', 'h', 'k^{-}', 'mu', 'theta', 'sigma_v', 'rho', 'sqrt(1-rho^2)',
          'lambda', 'mu_j', 'sigma_j^2')
    print(f"moment_y(l) returns a poly with keyfor = \n{kf}")
    args = sys.argv[1:]
    n = 1 if len(args) == 0 else int(args[0])
    print(f"moment_y({n}) = ")
    pprint(moment_y(n))
